/*
 * Scale Stack
 *
 * Copyright (C) 2010 Eric Day (eday@oddments.org)
 * All rights reserved.
 *
 * Use and distribution licensed under the BSD license. See the
 * COPYING file in the root project directory for full text.
 */

/**
 * @file
 * @brief Database Proxy Client Definitions
 */

#include "config.h"

#include <ScaleStack/DatabaseProxy/Client.h>
#include <ScaleStack/DatabaseProxy/ClientService.h>
#include <ScaleStack/DatabaseProxy/Connection.h>
#include <ScaleStack/Kernel/Module.h>

namespace ScaleStack
{

namespace DatabaseProxy
{

/*
 * Public methods.
 */

Client::Client(ClientService& client_service):
  SocketTCP::Connection(client_service),
  _state(),
  _next_state(),
  _need_yield(false),
  _shutdown(false),
  _size(),
  _packet_size(),
  _buffer(),
  _connection()
{
}

void Client::setConnection(DatabaseProxy::Connection* connection,
                           bool need_handshake)
{
  _connection = connection;

  if (connection == NULL)
    return;

  if (need_handshake && _state != SERVER_HANDSHAKE_PACKET &&
      _state != SERVER_HANDSHAKE_FLUSH)
  {
    _state = SERVER_HANDSHAKE_PACKET;
    reconnect();
  }
  else
    runNow();
}

void Client::yield(void)
{
  _need_yield = true;
}

void Client::connected(void)
{
  if (_connection == NULL)
    return;

  _connection->continueWrite();
}

size_t Client::read(uint8_t* buffer, size_t size)
{
  _size = size;
  _buffer = buffer;
  return _connectionWrite();
}

void Client::continueWrite(void)
{
  consume(_connectionWrite());
}

void Client::flushWrite(void)
{
  _connection->continueWrite();
}

void Client::error(void)
{
  _connection->error();
}

void Client::run(void)
{
  SocketTCP::Connection::run();

  if (_shutdown)
  {
    _connection->runNow();
    remove();
  }
  else
    _connection->continueWrite();
}

/*
 * Private methods.
 */

size_t Client::_connectionWrite(void)
{
  if (_connection == NULL)
    return 0;

  size_t written = 0;
  size_t total_written = 0;

  switch (_state)
  {
  case SERVER_HANDSHAKE_PACKET:
    if (_size < 4)
      return 0;

    _packet_size = 4 + (_buffer[0] | (static_cast<size_t>(_buffer[1]) << 8) | (static_cast<size_t>(_buffer[2]) << 16));
    _state = SERVER_HANDSHAKE_FLUSH;
    _module.logDebug("S-HAND packet: %u", _packet_size);

  case SERVER_HANDSHAKE_FLUSH:
    total_written = _connection->write(_buffer, _packet_size, true);
    _size -= total_written;
    _buffer += total_written;

    _packet_size -= total_written;
    if (_packet_size == 0)
      _state = RESULT_PACKET;

    break;

  case RESULT_PACKET:
    if (_size < 5)
      return 0;

    _packet_size = 4 + (_buffer[0] | (static_cast<size_t>(_buffer[1]) << 8) | (static_cast<size_t>(_buffer[2]) << 16));
    _state = RESULT_FLUSH;
    _module.logDebug("RESULT packet: %u", _packet_size);

    if (_buffer[4] == 0 || _buffer[4] >= 254)
      _next_state = RESULT_PACKET;
    else
      _next_state = COLUMN_PACKET;

  case RESULT_FLUSH:
    total_written = _connection->write(_buffer, _packet_size,
                                       _next_state == RESULT_PACKET);
    _size -= total_written;
    _buffer += total_written;

    _packet_size -= total_written;
    if (_packet_size == 0)
      _state = _next_state;

    if (_state != COLUMN_PACKET)
      break;

  case COLUMN_PACKET:
    while (1)
    {
      if (_size < 5)
        return 0;

      _packet_size = 4 + (_buffer[0] | (static_cast<size_t>(_buffer[1]) << 8) | (static_cast<size_t>(_buffer[2]) << 16));
      _state = COLUMN_FLUSH;
      _module.logDebug("COLUMN packet: %u", _packet_size);

      if (_packet_size == 9 && _buffer[4] == 254)
        _next_state = ROW_PACKET;
      else
        _next_state = COLUMN_PACKET;

  case COLUMN_FLUSH:
      written = _connection->write(_buffer, _packet_size, false);
      _size -= written;
      _buffer += written;
      total_written += written;

      _packet_size -= written;
      if (_packet_size == 0)
        _state = _next_state;

      if (_state == ROW_PACKET)
        break;
    }

  case ROW_PACKET:
    while (1)
    {
      if (_size < 5)
        return 0;

      _packet_size = 4 + (_buffer[0] | (static_cast<size_t>(_buffer[1]) << 8) | (static_cast<size_t>(_buffer[2]) << 16));
      _state = ROW_FLUSH;
      _module.logDebug("ROW packet: %u", _packet_size);

      if (_packet_size == 9 && _buffer[4] == 254)
        _next_state = RESULT_PACKET;
      else
        _next_state = ROW_PACKET;

  case ROW_FLUSH:
      written = _connection->write(_buffer, _packet_size,
                                   _next_state == RESULT_PACKET);
      _size -= written;
      _buffer += written;
      total_written += written;

      _packet_size -= written;
      if (_packet_size == 0)
        _state = _next_state;

      if (_state == RESULT_PACKET)
        break;
    }
  }

  if (_need_yield && _state == RESULT_PACKET && _packet_size == 0)
  {
    _need_yield = false;
    _connection->yield();
  }

  return total_written;
}

} /* namespace DatabaseProxy */

} /* namespace ScaleStack */
