/*
 * Scale Stack
 *
 * Copyright (C) 2010 Eric Day (eday@oddments.org)
 * All rights reserved.
 *
 * Use and distribution licensed under the BSD license. See the
 * COPYING file in the root project directory for full text.
 */

/**
 * @file
 * @brief Database Proxy Client Service Definitions
 */

#include "config.h"

#include <algorithm>

#include <ScaleStack/DatabaseProxy/Client.h>
#include <ScaleStack/DatabaseProxy/ClientService.h>
#include <ScaleStack/DatabaseProxy/Connection.h>
#include <ScaleStack/Kernel/Module.h>
#include <ScaleStack/Kernel/Option.h>

using namespace std;

namespace ScaleStack
{

namespace DatabaseProxy
{

/*
 * Public functions.
 */

ClientService::ClientService(Kernel::Module& module):
  SocketTCP::ConnectionService(module),
  _max_clients(_module.getOption("max_clients").getSizeValue()),
  _current_clients(),
  _queue_mutex(),
  _waiting_connections(),
  _free_clients()
{
  int return_code = pthread_mutex_init(&_queue_mutex, NULL);
  if (return_code != 0)
    _module.logFatal("Failed pthread_mutex_init: %d", return_code);
}

ClientService::~ClientService()
{
  shutdown();
  int return_code = pthread_mutex_destroy(&_queue_mutex);
  if (return_code != 0)
    _module.logError("Failed pthread_mutex_destroy: %d", return_code);
}

SocketTCP::Connection* ClientService::addConnection(void)
{
  return new Client(*this);
}

void ClientService::removeConnection(SocketTCP::Connection* connection)
{
  pthread_mutex_lock(&_queue_mutex);
  --_current_clients;
  pthread_mutex_unlock(&_queue_mutex);
  delete connection;
}

Client* ClientService::getClient(Connection* connection, bool need_handshake)
{
  pthread_mutex_lock(&_queue_mutex);

  Client* client;
  vector<Connection*>::reverse_iterator found_connection;
  found_connection = find(_waiting_connections.rbegin(),
                          _waiting_connections.rend(),
                          connection);

  if (_free_clients.empty() ||
      (_current_clients != _max_clients && need_handshake))
  {
    if (_current_clients == _max_clients)
    {
      if (found_connection == _waiting_connections.rend())
        _waiting_connections.insert(_waiting_connections.begin(), connection);
      pthread_mutex_unlock(&_queue_mutex);
      return NULL;
    }

    const string& host = _module.getOption("client_host").getValue();
    const string& port = _module.getOption("client_port").getValue();
    client = static_cast<Client*>(addConnect(host, port));
    ++_current_clients;
  }
  else
  {
    client = _free_clients.back();
    _free_clients.pop_back();
  }

  if (found_connection != _waiting_connections.rend())
    _waiting_connections.erase(found_connection.base() - 1);

  pthread_mutex_unlock(&_queue_mutex);

  client->setConnection(connection, need_handshake);
  return client;
}

void ClientService::yieldClient(Client* client)
{
  pthread_mutex_lock(&_queue_mutex);

  _free_clients.push_back(client);
  if (!_waiting_connections.empty())
    _waiting_connections.back()->runNow();

  pthread_mutex_unlock(&_queue_mutex);
}

void ClientService::removeFromQueue(Connection* connection)
{
  pthread_mutex_lock(&_queue_mutex);

  vector<Connection*>::iterator found_connection;
  found_connection = find(_waiting_connections.begin(),
                          _waiting_connections.end(),
                          connection);
  if (found_connection != _waiting_connections.end())
    _waiting_connections.erase(found_connection);

  pthread_mutex_unlock(&_queue_mutex);
}

} /* namespace DatabaseProxy */

} /* namespace ScaleStack */
