/*
 * Scale Stack
 *
 * Copyright (C) 2010 Eric Day (eday@oddments.org)
 * All rights reserved.
 *
 * Use and distribution licensed under the BSD license. See the
 * COPYING file in the root project directory for full text.
 */

/**
 * @file
 * @brief Database Proxy Connection Definitions
 */

#include "config.h"

#include <cstring>

#include <ScaleStack/DatabaseProxy/Client.h>
#include <ScaleStack/DatabaseProxy/Connection.h>
#include <ScaleStack/DatabaseProxy/Service.h>
#include <ScaleStack/Kernel/Module.h>

namespace ScaleStack
{

namespace DatabaseProxy
{

/*
 * Public methods.
 */

Connection::Connection(Service& service):
  SocketTCP::Connection(service),
  _state(),
  _need_handshake(true),
  _need_yield(false),
  _in_transaction(false),
  _shutdown(false),
  _size(),
  _packet_size(),
  _buffer(),
  _client()
{
  _client = service.getClient(this, _need_handshake);
}

size_t Connection::read(uint8_t* buffer, size_t size)
{
  _size = size;
  _buffer = buffer;
  return _clientWrite();
}

void Connection::continueWrite(void)
{
  consume(_clientWrite());
}

void Connection::flushWrite(void)
{
  if (_client != NULL)
    _client->continueWrite();
}

void Connection::error(void)
{
  _shutdown = true;
  runNow();
}

void Connection::run(void)
{
  if (_shutdown)
  {
    if (_client == NULL)
    {
      static_cast<Service&>(_service).removeFromQueue(this);
      remove();
      return;
    }

    static_cast<Service&>(_service).yieldClient(_client);
    _client = NULL;
    return;
  }

  if (_client == NULL)
    _client = static_cast<Service&>(_service).getClient(this, _need_handshake);
  else if (_need_yield)
  {
    _need_yield = false;
    static_cast<Service&>(_service).yieldClient(_client);
    _client = NULL;
  }
}

void Connection::yield(void)
{
  _need_yield = true;
  runNow();
}

/*
 * Private methods.
 */

size_t Connection::_clientWrite(void)
{
  if (_shutdown)
    return 0;

  if (_need_yield)
  {
    _need_yield = false;
    static_cast<Service&>(_service).yieldClient(_client);
    _client = NULL;
  }

  if (_client == NULL)
  {
    _client = static_cast<Service&>(_service).getClient(this, _need_handshake);
    /* Just return here, the client will call this when ready. */
    return 0;
  }

  size_t written = 0;

  switch (_state)
  {
  case CLIENT_HANDSHAKE_PACKET:
    if (_size < 4)
      return 0;

    _packet_size = 4 + (_buffer[0] | (static_cast<size_t>(_buffer[1]) << 8) | (static_cast<size_t>(_buffer[2]) << 16));
    _state = CLIENT_HANDSHAKE_FLUSH;
    _module.logDebug("C-HAND packet: %u", _packet_size);

  case CLIENT_HANDSHAKE_FLUSH:
    written = _client->write(_buffer, _packet_size, true);
    _size -= written;
    _buffer += written;

    _packet_size -= written;
    if (_packet_size == 0)
    {
      _state = COMMAND_PACKET;
      _need_handshake = false;
    }

    break;

  case COMMAND_PACKET:
    if (_size < 4)
      return 0;

    _packet_size = 4 + (_buffer[0] | (static_cast<size_t>(_buffer[1]) << 8) | (static_cast<size_t>(_buffer[2]) << 16));
    _state = COMMAND_FLUSH;
    _module.logDebug("COMMAND packet: %u", _packet_size);
    if (_packet_size == 10 && _size >= 10 && !memcmp("BEGIN", _buffer + 5, 5))
      _in_transaction = true;
    else if (!_in_transaction ||
             (_packet_size == 11 && _size >= 11 &&
              !memcmp("COMMIT", _buffer + 5, 6)))
    {
      _in_transaction = false;
    }

    if (_packet_size == 5 && _buffer[4] == 1)
    {
      error();
      return 5;
    }

  case COMMAND_FLUSH:
    written = _client->write(_buffer, _packet_size, true);
    _size -= written;
    _buffer += written;

    _packet_size -= written;
    if (_packet_size == 0)
      _state = COMMAND_PACKET;

    break;
  }

  if (!_in_transaction && _packet_size == 0)
    _client->yield();

  return written;
}

} /* namespace DatabaseProxy */

} /* namespace ScaleStack */
