/*
 * Scale Stack
 *
 * Copyright (C) 2010 Eric Day (eday@oddments.org)
 * All rights reserved.
 *
 * Use and distribution licensed under the BSD license. See the
 * COPYING file in the root project directory for full text.
 */

/**
 * @file
 * @brief Echo Server Test Module Definitions
 */

#include "config.h"

#include <cassert>
#include <cstdlib>
#include <cstring>

#include <ScaleStack/Kernel/Core.h>
#include <ScaleStack/Kernel/Module.h>
#include <ScaleStack/Kernel/Option.h>
#include <ScaleStack/SocketTCP/Connection.h>
#include <ScaleStack/SocketTCP/ConnectionService.h>

using namespace std;

static const size_t global_random_buffer_size = 8385637;
static uint8_t global_random_buffer[global_random_buffer_size];

namespace ScaleStack
{

namespace TestEchoServer
{

class Service: public SocketTCP::ConnectionService
{
public:
  Service(Kernel::Module& module);
  ~Service();
  SocketTCP::Connection* addConnection(void);
  void removeConnection(SocketTCP::Connection* connection);
};

class Connection: public SocketTCP::Connection
{
public:
  Connection(Service& service);
  void connected(void);
  size_t read(uint8_t* buffer, size_t size);
  void flushWrite(void);
  void error(void);
private:
  size_t _read_offset;
  size_t _write_offset;
};

/*
 * Public Service methods.
 */

Service::Service(Kernel::Module& module):
  SocketTCP::ConnectionService(module)
{
  for (int count = module.getOption("count").getIntValue();
       count != 0;
       --count)
  {
    addConnect(module.getOption("host").getValue(),
               module.getOption("port").getValue());
  }
}

Service::~Service()
{
  shutdown();
}

SocketTCP::Connection* Service::addConnection(void)
{
  return new Connection(*this);
}

void Service::removeConnection(SocketTCP::Connection* connection)
{
  delete connection;
}

/*
 * Public Connection methods.
 */

Connection::Connection(Service& service):
  SocketTCP::Connection(service),
  _read_offset(),
  _write_offset()
{
}

void Connection::connected(void)
{
  flushWrite();
}

size_t Connection::read(uint8_t* buffer, size_t size)
{
  assert(!memcmp(global_random_buffer + _read_offset, buffer, size));
  _read_offset += size;

  if (_read_offset == global_random_buffer_size)
  {
    if (_module.getOption("non_stop").getBoolValue())
    {
      _read_offset = 0;
      _write_offset = 0;

      if (_module.getOption("reconnect").getBoolValue())
        reconnect();
      else
        flushWrite();
    }
    else
      _module.getCore().shutdown();
  }

  return size;
}

void Connection::flushWrite(void)
{
  _write_offset += write(global_random_buffer + _write_offset,
                         global_random_buffer_size - _write_offset,
                         true);
}

void Connection::error(void)
{
  _module.getCore().shutdown();
  remove();
}

} /* namespace TestEchoServer */

} /* namespace ScaleStack */

using namespace ScaleStack;

static void options(Kernel::Module& module)
{
  module.addOption("host", "Host to test.", "HOST", "localhost");
  module.addOption("port", "Port to test.", "PORT", "32123");
  module.addOption("non_stop", "Run test forever.", "BOOLEAN", "False");
  module.addOption("reconnect", "Reconnect between tests.", "BOOLEAN", "False");
  module.addOption("count", "Number of connections to open.", "COUNT", "1");
}

static void start(Kernel::Module& module)
{
  for (size_t index = 0; index < global_random_buffer_size; index++)
    global_random_buffer[index] = static_cast<uint8_t>(rand());

  TestEchoServer::Service* service = new TestEchoServer::Service(module);
  try
  {
    module.getCore().addService("TestEchoServer", service);
  }
  catch (exception&)
  {
    delete service;
    throw;
  }
}

static void stop(Kernel::Module& module)
{
  delete module.getCore().removeService("TestEchoServer");
}

SCALESTACK_KERNEL_MODULE(options, start, stop, NULL, NULL);
