/*
 * Scale Stack
 *
 * Copyright (C) 2010 Eric Day (eday@oddments.org)
 * All rights reserved.
 *
 * Use and distribution licensed under the BSD license. See the
 * COPYING file in the root project directory for full text.
 */

/**
 * @file
 * @brief Event Handler and Handler Provider Declarations
 */

#ifndef SCALESTACK_EVENT_HANDLER_H
#define SCALESTACK_EVENT_HANDLER_H

#include <ctime>

#include <ScaleStack/Common/Macros.h>

namespace ScaleStack
{

namespace Event
{

class HandlerProvider;
class Service;

/**
 * This is the event handler base class. Other classes that want to add and
 * handle events must derive from this.
 */
class SCALESTACK_API Handler
{
  friend class HandlerProvider;

public:

  Handler(Service& service);

  virtual ~Handler();

  /**
   * Get the service this handler belongs to.
   */
  Service& getService(void);

  /**
   * This method is run after the event service has added it.
   */
  virtual void start(void);

  /**
   * Mark that this event handler should be removed.
   */
  virtual void remove(void);

  /**
   * This method is run when the handler encounters an error. The default
   * behavior is to call remove().
   */
  virtual void error(void);

  /**
   * Set the timer.
   *
   * @param[in] seconds The number of seconds from now for the timer to
   *  expire. A value of 0 cancels the timer.
   */
  virtual void setTimer(time_t seconds);

  /**
   * Run when the timer expires.
   */
  virtual void timerExpired(void);

  /**
   * Set the file descriptor to watch.
   *
   * @param[in] file_descriptor The file descriptor to use when watching
   *  for read and write events.
   */
  virtual void setFileDescriptor(int file_descriptor);

  /**
   * Watch the file descriptor for reading.
   */
  virtual void watchRead(void);

  /**
   * Run when the file descriptor is ready for reading.
   */
  virtual void readReady(int file_descriptor);

  /**
   * Watch the file descriptor for writing.
   */
  virtual void watchWrite(void);

  /**
   * Run when the file descriptor is ready for writing.
   */
  virtual void writeReady(int file_descriptor);

  /**
   * Notify the event service to run this handler now.
   */
  virtual void runNow(void);

  /**
   * Run when the event service receives a runNow() request.
   */
  virtual void run(void);

protected:

  Service& _service;

private:

  /**
   * Don't allow copying of objects.
   */
  SCALESTACK_LOCAL
  Handler(const Handler&);

  /**
   * Don't allow assignment of objects.
   */
  SCALESTACK_LOCAL
  Handler& operator=(const Handler&);

  HandlerProvider* _handler_provider;
};

/**
 * This is the base class for classes that provide event handlers. Instances
 * of this class are bound to Handler instances in order to provider the
 * implementation.
 */
class SCALESTACK_API HandlerProvider
{
public:

  HandlerProvider(void);

  virtual ~HandlerProvider();

  /**
   * Get the Handler bound to this provider.
   */
  Handler* getHandler(void) const;

  /**
   * Set the Handler for this HandlerProvider. This is used while initializing
   * the event handler in the service class. This is required in order
   * to reuse both Handler and HandlerProvider instances, which will be
   * implemented by unrelated classes.
   *
   * @param[in] handler Handler to set for this provider.
   */
  void setHandler(Handler* handler);

  /**
   * See Handler::start().
   */
  virtual void start(void);

  /**
   * See Handler::remove().
   */
  virtual void remove(void) = 0;

  /**
   * See Handler::error().
   */
  virtual void error(void);

  /**
   * See Handler::setTimer().
   */
  virtual void setTimer(time_t seconds) = 0;

  /**
   * See Handler::timerExpired().
   */
  virtual void timerExpired(void);

  /**
   * See Handler::setFileDescriptor().
   */
  virtual void setFileDescriptor(int file_descriptor) = 0;

  /**
   * See Handler::watchRead().
   */
  virtual void watchRead(void) = 0;

  /**
   * See Handler::readReady().
   */
  virtual void readReady(int file_descriptor);

  /**
   * See Handler::watchWrite().
   */
  virtual void watchWrite(void) = 0;

  /**
   * See Handler::writeReady().
   */
  virtual void writeReady(int file_descriptor);

  /**
   * See Handler::runNow().
   */
  virtual void runNow(void) = 0;

  /**
   * See Handler::run().
   */
  virtual void run(void);

private:

  /**
   * Don't allow copying of objects.
   */
  SCALESTACK_LOCAL
  HandlerProvider(const HandlerProvider&);

  /**
   * Don't allow assignment of objects.
   */
  SCALESTACK_LOCAL
  HandlerProvider& operator=(const HandlerProvider&);

  Handler* _handler;
};

/*
 * Public Handler methods.
 */

inline Handler::Handler(Service& service):
  _service(service),
  _handler_provider(NULL)
{
}

inline Handler::~Handler()
{
}

inline Service& Handler::getService(void)
{
  return _service;
}

inline void Handler::start(void)
{
}

inline void Handler::remove(void)
{
  _handler_provider->remove();
}

inline void Handler::error(void)
{
  remove();
}

inline void Handler::setTimer(time_t timer)
{
  _handler_provider->setTimer(timer);
}

inline void Handler::timerExpired(void)
{
}

inline void Handler::setFileDescriptor(int file_descriptor)
{
  _handler_provider->setFileDescriptor(file_descriptor);
}

inline void Handler::watchRead(void)
{
  _handler_provider->watchRead();
}

inline void Handler::readReady(int)
{
}

inline void Handler::watchWrite(void)
{
  _handler_provider->watchWrite();
}

inline void Handler::writeReady(int)
{
}

inline void Handler::runNow(void)
{
  _handler_provider->runNow();
}

inline void Handler::run(void)
{
}

/*
 * Public HandlerProvider methods.
 */

inline HandlerProvider::HandlerProvider(void):
  _handler(NULL)
{
}

inline HandlerProvider::~HandlerProvider()
{
}

inline Handler* HandlerProvider::getHandler(void) const
{
  return _handler;
}

inline void HandlerProvider::setHandler(Handler* handler)
{
  _handler = handler;
  _handler->_handler_provider = this;
}

inline void HandlerProvider::start(void)
{
  _handler->start();
}

inline void HandlerProvider::error(void)
{
  _handler->error();
}

inline void HandlerProvider::timerExpired(void)
{
  _handler->timerExpired();
}

inline void HandlerProvider::readReady(int file_descriptor)
{
  _handler->readReady(file_descriptor);
}

inline void HandlerProvider::writeReady(int file_descriptor)
{
  _handler->writeReady(file_descriptor);
}

inline void HandlerProvider::run(void)
{
  _handler->run();
}

} /* namespace Event */

} /* namespace ScaleStack */

#endif /* SCALESTACK_EVENT_HANDLER_H */
