/*
 * Scale Stack
 *
 * Copyright (C) 2010 Eric Day (eday@oddments.org)
 * All rights reserved.
 *
 * Use and distribution licensed under the BSD license. See the
 * COPYING file in the root project directory for full text.
 */

/**
 * @file
 * @brief Event Service Definitions
 */

#include "config.h"

#include <algorithm>
#include <exception>
#include <functional>

#include <ScaleStack/Event/Handler.h>
#include <ScaleStack/Event/Service.h>
#include <ScaleStack/Event/ServiceProvider.h>
#include <ScaleStack/Kernel/Core.h>
#include <ScaleStack/Kernel/Module.h>

using namespace std;

namespace ScaleStack
{

namespace Event
{

/*
 * Public methods.
 */

Service::Service(Kernel::Module& module):
  Kernel::Service(module),
  _shutdown(false),
  _service_provider(static_cast<ServiceProvider&>(module.getCore().getService("Event"))),
  _handler_mutex(),
  _shutdown_notification(),
  _handlers()
{
  int return_code = pthread_mutex_init(&_handler_mutex, NULL);
  if (return_code != 0)
    _module.logFatal("Failed pthread_mutex_init: %d", return_code);
}

Service::~Service()
{
  int return_code = pthread_mutex_destroy(&_handler_mutex);
  if (return_code != 0)
    _module.logError("Failed pthread_mutex_destroy: %d", return_code);
}

void Service::addHandler(Handler* handler)
{
  _service_provider.addHandler(handler);

  pthread_mutex_lock(&_handler_mutex);

  try
  {
    _handlers.insert(handler);
  }
  catch (exception&)
  {
    pthread_mutex_unlock(&_handler_mutex);
    throw;
  }

  pthread_mutex_unlock(&_handler_mutex);

  if (_shutdown)
    handler->remove();
}

void Service::shutdown(void)
{
  _shutdown = true;

  pthread_mutex_lock(&_handler_mutex);
  for_each(_handlers.begin(), _handlers.end(), mem_fun(&Handler::remove));

  while (!_handlers.empty())
  {
    pthread_mutex_unlock(&_handler_mutex);
    _shutdown_notification.wait();
    pthread_mutex_lock(&_handler_mutex);
  }

  pthread_mutex_unlock(&_handler_mutex);
}

/*
 * Private methods.
 */

void Service::_removeHandler(Handler* handler)
{
  removeHandler(handler);

  pthread_mutex_lock(&_handler_mutex);
  _handlers.erase(handler);
  pthread_mutex_unlock(&_handler_mutex);

  if (_shutdown)
    _shutdown_notification.send();
}

} /* namespace Event */

} /* namespace ScaleStack */
