/*
 * Scale Stack
 *
 * Copyright (C) 2010 Eric Day (eday@oddments.org)
 * All rights reserved.
 *
 * Use and distribution licensed under the BSD license. See the
 * COPYING file in the root project directory for full text.
 */

/**
 * @file
 * @brief Event Service Declarations
 */

#ifndef SCALESTACK_EVENT_SERVICE_H
#define SCALESTACK_EVENT_SERVICE_H

#include <pthread.h>
#include <set>

#include <ScaleStack/Common/Macros.h>
#include <ScaleStack/Common/Notification.h>
#include <ScaleStack/Kernel/Service.h>

namespace ScaleStack
{

namespace Event
{

class Handler;
class ServiceProvider;

/**
 * This is a base class that other classes using Handler classes must
 * implement. This provides a way to add and remove event handlers from the
 * event service provider.
 */
class SCALESTACK_API Service: public Kernel::Service
{
  friend class ServiceProvider;

public:

  /**
   * Constructor for event service provider.
   *
   * @param[in] module Module that the service is associated with.
   */
  Service(Kernel::Module& module);

  virtual ~Service();

  /**
   * Add an event handler to one of the event service threads.
   *
   * @param[in] handler Event handler to add.
   */
  virtual void addHandler(Handler* handler);

  /**
   * Remove an event handler. This is called by the event service when the
   * handler is no longer in use.
   *
   * @param[in] handler Event handler to remove.
   */
  virtual void removeHandler(Handler* handler) = 0;

  /**
   * Shutdown the service by removing all associated handlers. This can
   * block and should most likely only be used during shutdown in derived
   * class destructors.
   */
  virtual void shutdown(void);

private:

  /**
   * Don't allow copying of objects.
   */
  SCALESTACK_LOCAL
  Service(const Service&);

  /**
   * Don't allow assignment of objects.
   */
  SCALESTACK_LOCAL
  Service& operator=(const Service&);

  /**
   * Remove a handler from the service and call removeHandler().
   *
   * @param[in] handler Event handler to remove.
   */
  void _removeHandler(Handler* handler);

  bool _shutdown;
  ServiceProvider& _service_provider;
  pthread_mutex_t _handler_mutex;
  Common::Notification _shutdown_notification;
  std::set<Handler*> _handlers;
};

} /* namespace Event */

} /* namespace ScaleStack */

#endif /* SCALESTACK_EVENT_SERVICE_H */
