/*
 * Scale Stack
 *
 * Copyright (C) 2010 Eric Day (eday@oddments.org)
 * All rights reserved.
 *
 * Use and distribution licensed under the BSD license. See the
 * COPYING file in the root project directory for full text.
 */

/**
 * @file
 * @brief libevent Event Handler Provider Definitions
 */

#include "config.h"

#include <cerrno>
#include <fcntl.h>
#include <unistd.h>

#include <ScaleStack/EventLibevent/HandlerProvider.h>
#include <ScaleStack/EventLibevent/Thread.h>
#include <ScaleStack/Kernel/Module.h>

using namespace std;

namespace ScaleStack
{

namespace EventLibevent
{

/*
 * Public functions.
 */

extern "C" void HandlerProviderCallback(int file_descriptor, short events,
                                        void* context)
{
  HandlerProvider* handler_provider = static_cast<HandlerProvider*>(context);
  handler_provider->_callback(file_descriptor, events);
}

/*
 * Public methods.
 */

HandlerProvider::HandlerProvider(Kernel::Module& module):
  _in_check_queue(false),
  _need_start(false),
  _need_update(false),
  _need_run(false),
  _need_remove(false),
  _is_processing(false),
  _is_set(false),
  _is_added(false),
  _watch_read(false),
  _set_watch_read(false),
  _watch_write(false),
  _set_watch_write(false),
  _file_descriptor_updated(false),
  _file_descriptor(-1),
  _expire_time(0),
  _set_expire_time(0),
  _module(module),
  _thread(NULL),
  _event(),
  _timer()
{
}

HandlerProvider::~HandlerProvider()
{
  setThread(NULL);
}

void HandlerProvider::setThread(Thread* thread)
{
  _thread = thread;

  if (_thread == NULL)
  {
    if (_is_added)
    {
      event_del(&_event);
      _is_added = false;
    }
  }
  else
  {
    _need_start = true;
    if (!_in_check_queue)
      _thread->checkHandlerProvider(this);
  }
}

bool HandlerProvider::addToCheckQueue(void)
{
  if (_in_check_queue)
    return false;

  _module.logInfo("[%X] Adding to check queue", this);
  _in_check_queue = true;
  return true;
}

void HandlerProvider::check(void)
{
  _module.logInfo("[%X] Checking", this);

  _in_check_queue = false;
  _is_processing = true;

  if (_need_start)
  {
    _need_start = false;
    start();
  }

  if (_need_run)
  {
    _need_run = false;
    run();
  }

  _is_processing = false;
  _need_update = false;
  _updateEvent();
}

bool HandlerProvider::needRemove(void)
{
  return _need_remove;
}

void HandlerProvider::remove(void)
{
  _module.logInfo("[%X] Removing", this);
  _need_remove = true;
  if (!_in_check_queue)
    _thread->checkHandlerProvider(this);
}

void HandlerProvider::setTimer(time_t seconds)
{
  if (seconds == 0)
    _expire_time = 0;
  else
    _expire_time = _thread->getNow().tv_sec + seconds;

  _sendUpdateNotification();
}

void HandlerProvider::setFileDescriptor(int file_descriptor)
{
  _file_descriptor = file_descriptor;
  _file_descriptor_updated = true;

  if (_file_descriptor != -1)
  {
    int return_code = fcntl(_file_descriptor, F_GETFL, 0);
    if (return_code == -1)
    {
      close(_file_descriptor);
      _module.logFatal("Failed fcntl: %d", errno);
    }

    return_code = fcntl(_file_descriptor, F_SETFL, return_code | O_NONBLOCK);
    if (return_code == -1)
    {
      close(_file_descriptor);
      _module.logFatal("Failed fcntl: %d", errno);
    }
  }

  _sendUpdateNotification();
}

void HandlerProvider::watchRead(void)
{
  _watch_read = true;
  _sendUpdateNotification();
}

void HandlerProvider::watchWrite(void)
{
  _watch_write = true;
  _sendUpdateNotification();
}

void HandlerProvider::runNow(void)
{
  _need_run = true;
  if (!_in_check_queue)
    _thread->checkHandlerProvider(this);
}

/*
 * Private methods.
 */

void HandlerProvider::_sendUpdateNotification(void)
{
  /*
   * We call updateEvent() after processing callbacks, so don't bother doing
   * anything now if we're already in callbacks. This may be due to recursion
   * or coincidence with another thread.
   */
  if (_is_processing)
    return;

  /* Otherwise, notify the thread that this handler has changed. */
  if (_set_expire_time > _expire_time ||
      _file_descriptor_updated ||
      _set_watch_read != _watch_read ||
      _set_watch_write != _watch_write)
  {
    _need_update = true;
    if (!_in_check_queue)
      _thread->checkHandlerProvider(this);
  }
}

void HandlerProvider::_updateEvent(void)
{
  _need_update = false;

  if (_need_start)
    return;

  if (_set_expire_time <= _expire_time &&
      !_file_descriptor_updated &&
      _set_watch_read == _watch_read &&
      _set_watch_write == _watch_write)
  {
    return;
  }

  if (_is_added)
  {
    event_del(&_event);
    _is_added = false;
  }

  if (!_is_set ||
      _file_descriptor_updated ||
      _set_watch_read != _watch_read ||
      _set_watch_write != _watch_write)
  {
    short events = EV_PERSIST;

    if (_file_descriptor != -1)
    {
      if (_watch_read)
        events |= EV_READ;
      if (_watch_write)
        events |= EV_WRITE;
    }

    event_set(&_event, _file_descriptor, events, HandlerProviderCallback, this);

    if (event_base_set(_thread->getEventBase(), &_event) == -1)
    {
      _is_set = false;
      _module.logError("Failed event_base_set: -1");
      remove();
      return;
    }

    _is_set = true;
    _file_descriptor_updated = false;
    _set_watch_read = _watch_read;
    _set_watch_write = _watch_write;
  }

  int return_code;
  if (_expire_time > 0)
  {
    _timer.tv_sec = _expire_time;
    return_code = event_add(&_event, &_timer);
  }
  else if (_file_descriptor != -1)
    return_code = event_add(&_event, NULL);
  else
    return_code = 0;

  if (return_code == -1)
  {
    _is_added = false;
    _module.logError("Failed event_add: -1");
    remove();
    return;
  }

  _is_added = true;
}

void HandlerProvider::_callback(int file_descriptor, short events)
{
  _is_processing = true;

  if (events & EV_TIMEOUT && !_need_remove)
  {
    _set_expire_time = 0;

    if (_thread->getNow().tv_sec >= _expire_time)
    {
      _module.logDebug("[%X] Timer expired", this);
      _expire_time = 0;
      timerExpired();
    }
  }

  if (events & EV_READ && !_need_remove)
  {
    _module.logDebug("[%X] Read ready", this);
    _watch_read = false;
    readReady(file_descriptor);
  }

  if (events & EV_WRITE && !_need_remove)
  {
    _module.logDebug("[%X] Write ready", this);
    _watch_write = false;
    writeReady(file_descriptor);
  }

  _is_processing = false;

  _updateEvent();
}

} /* namespace EventLibevent */

} /* namespace ScaleStack */
