/*
 * Scale Stack
 *
 * Copyright (C) 2010 Eric Day (eday@oddments.org)
 * All rights reserved.
 *
 * Use and distribution licensed under the BSD license. See the
 * COPYING file in the root project directory for full text.
 */

/**
 * @file
 * @brief libevent Event Handler Provider Declarations
 */

#ifndef SCALESTACK_EVENTLIBEVENT_HANDLERPROVIDER_H
#define SCALESTACK_EVENTLIBEVENT_HANDLERPROVIDER_H

#include <event.h>

#include <ScaleStack/Common/Macros.h>
#include <ScaleStack/Event/Handler.h>

namespace ScaleStack
{

namespace Kernel
{

class Module;

} /* namespace Kernel */

namespace EventLibevent
{

class Thread;

/**
 * Function passed to libevent to invoke HandlerProvider::_callback().
 */
extern "C" void HandlerProviderCallback(int file_descriptor, short events,
                                        void* context);

/**
 * This is the HandlerProvider implementation using libevent.
 */
class SCALESTACK_API HandlerProvider: public Event::HandlerProvider
{
  friend void HandlerProviderCallback(int file_descriptor, short events,
                                      void* context);

public:

  /**
   * Constructor for handler provider.
   *
   * @param[in] module Module that the handler is associated with.
   */
  HandlerProvider(Kernel::Module& module);

  ~HandlerProvider();

  /**
   * Set the thread this handler provider belongs to.
   *
   * @param[in] thread Thread to set.
   */
  void setThread(Thread* thread);

  /**
   * Ask if the the handler provider should be added to the check queue. If
   * so, also mark that it is now in the queue.
   */
  bool addToCheckQueue(void);

  /**
   * Check to see if handler provider needs to be started, ran, or updated.
   */
  void check(void);

  /**
   * Check to see if this handler provider should be removed.
   */
  bool needRemove(void);

  /**
   * See Event::Handler::remove().
   */
  void remove(void);

  /**
   * See Event::Handler::setTimer().
   */
  void setTimer(time_t seconds);

  /**
   * See Event::Handler::setFileDescriptor().
   */
  void setFileDescriptor(int file_descriptor);

  /**
   * See Event::Handler::watchRead().
   */
  void watchRead(void);

  /**
   * See Event::Handler::watchWrite().
   */
  void watchWrite(void);

  /**
   * See Event::Handler::runNow().
   */
  void runNow(void);

private:

  /**
   * Don't allow copying of objects.
   */
  SCALESTACK_LOCAL
  HandlerProvider(const HandlerProvider&);

  /**
   * Update the event handler structure with any changes.
   */
  SCALESTACK_LOCAL
  void _updateEvent(void);

  /**
   * Don't allow assignment of objects.
   */
  SCALESTACK_LOCAL
  HandlerProvider& operator=(const HandlerProvider&);

  /**
   * Notify handler provider that events should be updated.
   */
  SCALESTACK_LOCAL
  void _sendUpdateNotification(void);

  /**
   * This is called when a libevent event notification happens.
   *
   * @param[in] file_descriptor File descriptor for the event.
   * @param[in] events Type of events that are ready.
   */
  SCALESTACK_LOCAL
  void _callback(int file_descriptor, short events);

  bool _in_check_queue;
  bool _need_start;
  bool _need_update;
  bool _need_run;
  bool _need_remove;
  bool _is_processing;
  bool _is_set;
  bool _is_added;
  bool _watch_read;
  bool _set_watch_read;
  bool _watch_write;
  bool _set_watch_write;
  int _file_descriptor_updated;
  int _file_descriptor;
  time_t _expire_time;
  time_t _set_expire_time;
  Kernel::Module& _module;
  Thread* _thread;
  struct event _event;
  struct timeval _timer;
};

} /* namespace EventLibevent */

} /* namespace ScaleStack */

#endif /* SCALESTACK_EVENTLIBEVENT_HANDLERPROVIDER_H */
