/*
 * Scale Stack
 *
 * Copyright (C) 2010 Eric Day (eday@oddments.org)
 * All rights reserved.
 *
 * Use and distribution licensed under the BSD license. See the
 * COPYING file in the root project directory for full text.
 */

/**
 * @file
 * @brief libevent Event Service Definitions
 */

#include "config.h"

#include <algorithm>
#include <exception>
#include <functional>

#include <ScaleStack/Event/Service.h>
#include <ScaleStack/EventLibevent/HandlerProvider.h>
#include <ScaleStack/EventLibevent/ServiceProvider.h>
#include <ScaleStack/EventLibevent/Thread.h>
#include <ScaleStack/Kernel/Module.h>
#include <ScaleStack/Kernel/Option.h>

using namespace std;

namespace ScaleStack
{

namespace EventLibevent
{

/*
 * Public methods.
 */

ServiceProvider::ServiceProvider(Kernel::Module& module):
  Event::ServiceProvider(module),
  _threads(),
  _current_thread()
{
  size_t thread_count = _module.getOption("threads").getSizeValue();
  if (thread_count == 0)
    _module.logFatal("The 'threads' option must be greater than 0");

  for (size_t thread_index = 0; thread_index < thread_count; thread_index++)
  {
    Thread* thread = NULL;
    try
    {
      thread = new Thread(*this, thread_index);
      _threads.push_back(thread);
    }
    catch (exception&)
    {
      if (thread != NULL)
      {
        thread->stop();
        delete thread;
      }

      _removeThreads();
      throw;
    }
  }

  _current_thread = _threads.begin();
}

ServiceProvider::~ServiceProvider()
{
  _removeThreads();
}

void ServiceProvider::addHandler(Event::Handler* handler)
{
  HandlerProvider* handler_provider = new HandlerProvider(_module);
  try
  {
    handler_provider->setHandler(handler);

    /* Choose assigned thread by round-robin. */
    if (_current_thread == _threads.end())
      _current_thread = _threads.begin();

    handler_provider->setThread(*(_current_thread++));
  }
  catch (exception&)
  {
    delete handler_provider;
    throw;
  }
}

void ServiceProvider::removeHandlerProvider(HandlerProvider* handler_provider)
{
  removeHandler(handler_provider->getHandler());
  delete handler_provider;
}

/*
 * Private methods.
 */

void ServiceProvider::_removeThreads(void)
{
  for_each(_threads.begin(), _threads.end(), mem_fun(&Thread::stop));

  for (_current_thread = _threads.begin();
       _current_thread != _threads.end();
       ++_current_thread)
  {
    delete *_current_thread;
  }
}

} /* namespace EventLibevent */

} /* namespace ScaleStack */
