/*
 * Scale Stack
 *
 * Copyright (C) 2010 Eric Day (eday@oddments.org)
 * All rights reserved.
 *
 * Use and distribution licensed under the BSD license. See the
 * COPYING file in the root project directory for full text.
 */

/**
 * @file
 * @brief libevent Event Thread Definitions
 */

#include "config.h"

#include <algorithm>
#include <cerrno>
#include <fcntl.h>
#include <functional>
#include <unistd.h>

#include <ScaleStack/EventLibevent/HandlerProvider.h>
#include <ScaleStack/EventLibevent/ServiceProvider.h>
#include <ScaleStack/EventLibevent/Thread.h>
#include <ScaleStack/Kernel/Core.h>
#include <ScaleStack/Kernel/Module.h>
#include <ScaleStack/Kernel/Option.h>

using namespace std;

namespace ScaleStack
{

namespace EventLibevent
{

/*
 * Public functions.
 */

extern "C" void* ThreadRun(void* context)
{
  Thread* thread = static_cast<Thread*>(context);
  thread->_run();
  return NULL;
}

extern "C" void ThreadHandleNotification(int file_descriptor, short events,
                                         void* context)
{
  Thread* thread = static_cast<Thread*>(context);
  thread->_handleNotification(file_descriptor, events);
}

/*
 * Public methods.
 */

Thread::Thread(ServiceProvider& service_provider, size_t thread_index):
  _shutdown(false),
  _handler_providers_index(),
  _thread_index(thread_index),
  _module(service_provider.getModule()),
  _service_provider(service_provider),
  _base(),
  _notification_pipe(),
  _thread_id(),
  _notification_event(),
  _notification_timeout(),
  _now(),
  _handler_providers_mutex(),
  _handler_providers()
{
  if (gettimeofday(&_now, 0) == -1)
    _module.logFatal("Failed gettimeofday: %d", errno);

  _base = event_base_new();
  if (_base == NULL)
    _module.logFatal("Failed event_base_new: NULL");

  _module.logDebug("[Thread %u] Using libevent %s with %s", _thread_index,
                   event_get_version(), event_base_get_method(_base));

  try
  {
    _addMutexes();

    try
    {
      _addNotificationPipe();

      try
      {
        _addThread();
      }
      catch (exception&)
      {
        _removeNotificationPipe();
        throw;
      }
    }
    catch (exception&)
    {
      _removeMutexes();
      throw;
    }
  }
  catch (exception&)
  {
    event_base_free(_base);
    throw;
  }

  _module.logInfo("[Thread %u] Constructed", _thread_index);
}

Thread::~Thread()
{
  _removeThread();
  _removeNotificationPipe();
  _removeMutexes();
  event_base_free(_base);
  _module.logInfo("[Thread %u] Destructed", _thread_index);
}

void Thread::stop(void)
{
  _module.logInfo("[Thread %u] Stopping", _thread_index);
  _shutdown = true;
  _sendNotification();
}

void Thread::checkHandlerProvider(HandlerProvider* handler_provider)
{
  pthread_mutex_lock(&_handler_providers_mutex);

  /* Double check queue flag inside lock to avoid race conditions. */
  if (handler_provider->addToCheckQueue())
    _handler_providers[_handler_providers_index].push_back(handler_provider);

  pthread_mutex_unlock(&_handler_providers_mutex);

  _sendNotification();
}

struct event_base* Thread::getEventBase(void)
{
  return _base;
}

const struct timeval& Thread::getNow(void)
{
  return _now;
}

/*
 * Private methods.
 */

void Thread::_run(void)
{
  _module.logInfo("[Thread %u] Running", _thread_index);

  if (event_base_loop(_base, 0) == -1)
  {
    _module.logError("[Thread %u] event_base_loop: -1", _thread_index);
    _module.getCore().shutdown();
  }

  _module.logInfo("[Thread %u] Stopped", _thread_index);
}

void Thread::_addNotificationPipe(void)
{
  int return_code = pipe(_notification_pipe);
  if (return_code == -1)
    _module.logFatal("Failed pipe: %d", errno);

  for (int index = 0; index < 2; index++)
  {
    return_code = fcntl(_notification_pipe[index], F_GETFL, 0);
    if (return_code == -1)
    {
      _removeNotificationPipe();
      _module.logFatal("Failed fcntl: %d", errno);
    }

    return_code = fcntl(_notification_pipe[index], F_SETFL, return_code | O_NONBLOCK);
    if (return_code == -1)
    {
      _removeNotificationPipe();
      _module.logFatal("Failed fcntl: %d", errno);
    }
  }

  event_set(&_notification_event, _notification_pipe[0], EV_READ | EV_PERSIST,
            ThreadHandleNotification, this);

  if (event_base_set(_base, &_notification_event) == -1)
  {
    _removeNotificationPipe();
    _module.logFatal("Failed event_base_set: -1");
  }

  _notification_timeout.tv_sec = 1;
  _notification_timeout.tv_usec = 0;
  if (event_add(&_notification_event, &_notification_timeout) == -1)
  {
    _removeNotificationPipe();
    _module.logFatal("Failed event_add: -1");
  }
}

void Thread::_removeNotificationPipe(void)
{
  close(_notification_pipe[0]);
  close(_notification_pipe[1]);
}

void Thread::_sendNotification(void)
{
  if (write(_notification_pipe[1], "\0", 1) == -1)
    _module.logError("Could not write to notification pipe: %d", errno);
}

void Thread::_handleNotification(int file_descriptor, short events)
{
  _module.logDebug("[Thread %u] Notified", _thread_index);

  if (events & EV_READ)
  {
    char buffer[1024];
    if (read(file_descriptor, buffer, sizeof(buffer)) == -1)
      _module.logError("Could not read from notification pipe: %d", errno);
  }

  if (event_del(&_notification_event) == -1)
  {
    /* Not much we can really do here, but this should never happen. */
    _module.logError("Failed event_del: -1");
    _module.getCore().shutdown();
  }
  else if (!_shutdown)
  {
    if (event_add(&_notification_event, &_notification_timeout) == -1)
    {
      /* Not much we can really do here, but this should never happen. */
      _module.logError("Failed event_add: -1");
      _module.getCore().shutdown();
    }
  }

  if (gettimeofday(&_now, 0) == -1)
    _module.logError("Failed gettimeofday: %d", errno);

  /* Just switch the index, keep lock time to a minimum. */
  uint8_t index = _handler_providers_index;
  pthread_mutex_lock(&_handler_providers_mutex);
  _handler_providers_index = static_cast<uint8_t>(1 - _handler_providers_index);
  pthread_mutex_unlock(&_handler_providers_mutex);

  vector<HandlerProvider*>::iterator handler_provider;
  for (handler_provider = _handler_providers[index].begin();
       handler_provider != _handler_providers[index].end();
       ++handler_provider)
  {
    if ((*handler_provider)->needRemove())
      _service_provider.removeHandlerProvider(*handler_provider);
    else
      (*handler_provider)->check();
  }

  _handler_providers[index].clear();
}

void Thread::_addMutexes(void)
{
  int return_code = pthread_mutex_init(&_handler_providers_mutex, NULL);
  if (return_code != 0)
    _module.logFatal("Failed pthread_mutex_init: %d", return_code);
}

void Thread::_removeMutexes(void)
{
  int return_code = pthread_mutex_destroy(&_handler_providers_mutex);
  if (return_code != 0)
    _module.logError("Failed pthread_mutex_destroy: %d", return_code);
}

void Thread::_addThread(void)
{
  pthread_attr_t thread_attribute;

  int return_code = pthread_attr_init(&thread_attribute);
  if (return_code != 0)
    _module.logFatal("Failed pthread_attr_init: %d", return_code);

  size_t stack_size = _module.getOption("thread_stack_size").getSizeValue();
  if (stack_size != 0)
  {
    return_code = pthread_attr_setstacksize(&thread_attribute, stack_size);
    if (return_code != 0)
    {
      (void) pthread_attr_destroy(&thread_attribute);
      _module.logFatal("Failed pthread_attr_setstacksize: %d", return_code);
    }
  }

  return_code = pthread_attr_setscope(&thread_attribute, PTHREAD_SCOPE_SYSTEM);
  if (return_code != 0)
  {
    (void) pthread_attr_destroy(&thread_attribute);
    _module.logFatal("Failed pthread_attr_setscope: %d", return_code);
  }

  return_code = pthread_create(&_thread_id, &thread_attribute, ThreadRun, this);
  if (return_code != 0)
  {
    (void) pthread_attr_destroy(&thread_attribute);
    _module.logFatal("Failed pthread_create: %d", return_code);
  }

  return_code = pthread_attr_destroy(&thread_attribute);
  if (return_code != 0)
    _module.logError("Failed pthread_attr_destroy: %d", return_code);
}

void Thread::_removeThread(void)
{
  int return_code = pthread_join(_thread_id, NULL);
  if (return_code != 0)
    _module.logError("Failed pthread_join: %d", return_code);
}

} /* namespace EventLibevent */

} /* namespace ScaleStack */
