/*
 * Scale Stack
 *
 * Copyright (C) 2010 Eric Day (eday@oddments.org)
 * All rights reserved.
 *
 * Use and distribution licensed under the BSD license. See the
 * COPYING file in the root project directory for full text.
 */

/**
 * @file
 * @brief libevent Event Thread Declarations
 */

#ifndef SCALESTACK_EVENTLIBEVENT_THREAD_H
#define SCALESTACK_EVENTLIBEVENT_THREAD_H

#include <event.h>
#include <pthread.h>
#include <stdint.h>
#include <vector>

#include <ScaleStack/Common/Macros.h>

namespace ScaleStack
{

namespace Kernel
{

class Module;

} /* namespace Kernel */

namespace EventLibevent
{

class HandlerProvider;
class ServiceProvider;

/**
 * Function passed to pthread_create() to invoke Thread::_run().
 */
extern "C" void* ThreadRun(void* context);

/**
 * Function used by notification event to invoke Thread::_handleNotification().
 */
extern "C" void ThreadHandleNotification(int file_descriptor, short events,
                                         void* context);

class Thread
{
  friend void* ThreadRun(void*);
  friend void ThreadHandleNotification(int file_descriptor, short events,
                                       void* context);

public:

  /**
   * Constructor for event threads. This also starts a new pthread. Once
   * starting this call, the _run() method can be called at any time, even
   * before the constructor completes.
   *
   * @param[in] service_provider Service provider creating this thread.
   * @param[in] thread_index The 0-based index for this thread.
   */
  Thread(ServiceProvider& service_provider, size_t thread_index);

  ~Thread();

  /**
   * Notify a thread to stop running. This is called by the service provider
   * class for each thread before deleting the threads. This allows all
   * threads to be notified to stop before any joins are attempted.
   */
  void stop(void);

  /**
   * Queue a handler provider to be checked on this thread.
   *
   * @param[in] handler_provider Event handler provider to check.
   */
  void checkHandlerProvider(HandlerProvider* handler_provider);

  /**
   * Get the event base struct. This is used by handler providers once they
   * are added to this thread.
   */
  struct event_base* getEventBase(void);

  /**
   * Get a cached time structure.
   */
  const struct timeval& getNow(void);

private:

  /**
   * Don't allow copying of objects.
   */
  SCALESTACK_LOCAL
  Thread(const Thread&);

  /**
   * Don't allow assignment of objects.
   */
  SCALESTACK_LOCAL
  Thread& operator=(const Thread&);

  /**
   * Method that implements the body of the running thread.
   */
  SCALESTACK_LOCAL
  void _run(void);

  /**
   * Start the notification pipe.
   */
  SCALESTACK_LOCAL
  void _addNotificationPipe(void);

  /**
   * Stop the notification pipe.
   */
  SCALESTACK_LOCAL
  void _removeNotificationPipe(void);

  /**
   * Send a notification to run and process events.
   */
  SCALESTACK_LOCAL
  void _sendNotification(void);

  /**
   * Handle a notification event.
   */
  SCALESTACK_LOCAL
  void _handleNotification(int file_descriptor, short events);

  /**
   * Add mutexes needed for thread.
   */
  SCALESTACK_LOCAL
  void _addMutexes(void);

  /**
   * Remove mutexes added in _addMutexes().
   */
  SCALESTACK_LOCAL
  void _removeMutexes(void);

  /**
   * Start the pthread.
   */
  SCALESTACK_LOCAL
  void _addThread(void);

  /**
   * Stop the pthread by joining it.
   */
  SCALESTACK_LOCAL
  void _removeThread(void);

  bool _shutdown;
  uint8_t _handler_providers_index;
  size_t _thread_index;
  Kernel::Module& _module;
  ServiceProvider& _service_provider;
  struct event_base* _base;
  int _notification_pipe[2];
  pthread_t _thread_id;
  struct event _notification_event;
  struct timeval _notification_timeout;
  struct timeval _now;
  pthread_mutex_t _handler_providers_mutex;
  std::vector<HandlerProvider*> _handler_providers[2];
};

} /* namespace EventLibevent */

} /* namespace ScaleStack */

#endif /* SCALESTACK_EVENTLIBEVENT_THREAD_H */
