/*
 * Scale Stack
 *
 * Copyright (C) 2010 Eric Day (eday@oddments.org)
 * All rights reserved.
 *
 * Use and distribution licensed under the BSD license. See the
 * COPYING file in the root project directory for full text.
 */

/**
 * @file
 * @brief Kernel Core Definitions
 */

#include "config.h"

#include <algorithm>
#include <cerrno>
#include <cstring>
#include <exception>
#include <functional>
#include <iostream>
#include <signal.h>
#include <strings.h>
#include <unistd.h>

#include <ScaleStack/Common/Exception.h>
#include <ScaleStack/Common/FunctionObjects.h>
#include <ScaleStack/Kernel/Core.h>
#include <ScaleStack/Kernel/LogConsole.h>
#include <ScaleStack/Kernel/Module.h>

using namespace std;

namespace ScaleStack
{

namespace Kernel
{

/*
 * This is provided for convenience for applications that only use a single
 * Core instance per process. This needs to be 'extern "C"' and not just
 * static to make SunCC happy.
 */

static ScaleStack::Kernel::Core* global_kernel_core = NULL;
extern "C" void global_signal_handler(int signal_number);
extern "C" void global_signal_handler(int signal_number)
{
  global_kernel_core->handleSignal(signal_number);
}

/*
 * Public methods.
 */

Core::Core(void):
  LogMessage(0),
  _print_help(false),
  _shutdown_request(false),
  _iterator_on_modules(false),
  _run_notification(),
  _paths(),
  _modules(),
  _add_modules(),
  _moduleIterator(_modules.begin()),
  _aliases(),
  _services(),
  _logs()
{
  _paths.push_back(SCALESTACK_QUOTE_VALUE(SCALESTACK_MODULE_PATH));

  Log* log = new LogConsole();
  try
  {
    addLog("Console", log);
  }
  catch (exception&)
  {
    delete log;
    throw;
  }
}

Core::~Core()
{
  logInfo("Destructing core object");

  for_each(_add_modules.begin(), _add_modules.end(), mem_fun(&Module::_unload));
  for_each(_modules.begin(), _modules.end(), mem_fun(&Module::_unload));
  for_each(_add_modules.begin(), _add_modules.end(), Common::DeletePointer());
  for_each(_modules.begin(), _modules.end(), Common::DeletePointer());

  if (getServiceCount() != 0)
  {
    logError("Services still exist");

    map<string, Service*>::const_iterator service;
    for (service = _services.begin(); service != _services.end(); ++service)
      logError("  %s", service->first.c_str());
  }

  /* There should only be 1 left for LogConsole. */
  if (getLogCount() != 1)
  {
    logError("Logs still exist");

    map<string, Log*>::const_iterator log;
    for (log = _logs.begin(); log != _logs.end(); ++log)
      logError("  %s", log->first.c_str());
  }

  logInfo("Destructed core object");

  delete removeLog("Console");
}

void Core::parseArguments(int argument_count, const char** arguments)
{
  /*
   * Do one scan through arguments to find verbose flags first. We want to
   * know the full verbosity level before actually parsing real entries,
   * and verbose options may not be specified until the end.
   */
  for (int current = 0; current < argument_count; current++)
  {
    string argument(arguments[current]);

    size_t start_module = argument.find_first_not_of("-");
    if (start_module == string::npos)
      logFatal("Bad command line option: %s", argument.c_str());
    else if (start_module > 0)
      argument.erase(0, start_module);

    /* Check for verbose flags. */
    if (argument.find_first_not_of("v") == string::npos)
      _log_threshold += argument.size();
  }

  for (int current = 0; current < argument_count; current++)
  {
    string argument(arguments[current]);
    size_t start_module = argument.find_first_not_of("-");
    if (start_module > 0)
      argument.erase(0, start_module);

    /* Skip verbose flags. */
    if (argument.find_first_not_of("v") == string::npos)
      continue;

    /* Check for help flags. */
    if (argument == "help" || argument == "h")
    {
      _print_help = true;
      continue;
    }

    /* Parse as a normal option. */
    size_t start_option = argument.find(".");
    size_t start_value = argument.find("=");
    if (start_option > start_value)
      start_option = string::npos;

    string module_name;
    if (start_option == string::npos)
    {
      if (start_value == string::npos)
        module_name = argument;
      else
        module_name = string(argument, 0, start_value);
    }
    else
      module_name = string(argument, 0, start_option);

    string option_name;
    if (start_option != string::npos)
    {
      start_option++;
      if (start_value == string::npos)
        option_name = string(argument, start_option);
      else
        option_name = string(argument, start_option, start_value - start_option);
    }

    string value;
    if (start_value != string::npos)
    {
      start_value++;
      value = string(argument, start_value);
    }

    addModuleOptionValue(module_name, option_name, value);
  }
}

void Core::addModuleOptionValue(const std::string& module_name,
                                const std::string& option_name,
                                const std::string& value)
{
  logDebug("Setting option: Module=%s Option=%s Value=%s",
           module_name.c_str(), option_name.c_str(), value.c_str());

  if (module_name == "")
    logFatal("Invalid module name: %s", module_name.c_str());

  /* Check for additional search paths. */
  if (module_name == "path" || module_name == "p")
  {
    if (option_name != "" || value == "")
      logFatal("Invalid path option: %s", module_name.c_str());

    _paths.insert(_paths.begin(), value);
    return;
  }

  if (option_name == "")
  {
    if (value == "")
    {
      /* Only module name given, add it to the list of modules to load. */
      getOrAddModule(module_name);
      return;
    }

    /* Module name and value given, add an alias for the module name. */
    map<string, string>::iterator alias;
    for (alias = _aliases.begin(); alias != _aliases.end(); ++alias)
    {
      if (!strcasecmp(alias->first.c_str(), module_name.c_str()))
      {
        alias->second = value;
        break;
      }
    }

    if (alias == _aliases.end())
      _aliases.insert(pair<string, string>(module_name, value));

    logInfo("Added module alias: %s=%s", module_name.c_str(), value.c_str());
    return;
  }

  getOrAddModule(module_name)._setOptionValue(option_name, value);
}

void Core::setLogThreshold(size_t log_threshold)
{
  _log_threshold = log_threshold;
}

void Core::setSignalHandlers(void)
{
  if (global_kernel_core != NULL)
    logFatal("Global kernel core already set for signal handling");

  logInfo("Setting signal handlers");

  global_kernel_core = this;

  struct sigaction signal_action;
  memset(&signal_action, 0, sizeof(struct sigaction));
  signal_action.sa_handler = SIG_IGN;

  if (sigemptyset(&signal_action.sa_mask) == -1 ||
      sigaction(SIGPIPE, &signal_action, 0) == -1)
  {
    logFatal("Could not set SIGPIPE handler: %d", errno);
  }

  signal_action.sa_handler = global_signal_handler;

  if (sigaction(SIGHUP, &signal_action, 0) == -1)
    logFatal("Could not set SIGHUP handler: %d", errno);

  if (sigaction(SIGINT, &signal_action, 0) == -1)
    logFatal("Could not set SIGINT handler: %d", errno);

  if (sigaction(SIGQUIT, &signal_action, 0) == -1)
    logFatal("Could not set SIGQUIT handler: %d", errno);

  if (sigaction(SIGTERM, &signal_action, 0) == -1)
    logFatal("Could not set SIGTERM handler: %d", errno);

  if (sigaction(SIGUSR1, &signal_action, 0) == -1)
    logFatal("Could not set SIGUSR1 handler: %d", errno);

  if (sigaction(SIGUSR2, &signal_action, 0) == -1)
    logFatal("Could not set SIGUSR2 handler: %d", errno);
}

void Core::handleSignal(int signal_number)
{
  logInfo("Caught signal: %d", signal_number);

  if (signal_number == SIGQUIT ||
      signal_number == SIGTERM ||
      signal_number == SIGINT)
  {
    shutdown();
  }
}

Module& Core::getOrAddModule(const std::string& name)
{
  const string& alias = _getAlias(name);

  Module* module = _getModule(alias);
  if (module != NULL)
    return *module;

  module = new Module(*this, alias, _log_threshold);
  try
  {
    _add_modules.push_back(module);
  }
  catch (exception&)
  {
    delete module;
    throw;
  }

  return *module;
}

Module& Core::getModule(const std::string& name) const
{
  const string& alias = _getAlias(name);

  Module* module = _getModule(alias);
  if (module == NULL)
    logFatal("Could not find module: %s", alias.c_str());

  return *module;
}

size_t Core::getModuleCount(void) const
{
  return _modules.size() + _add_modules.size();
}

Module* Core::beginModule(void)
{
  _moduleIterator = _modules.begin();
  _iterator_on_modules = true;
  return nextModule();
}

Module* Core::nextModule(void)
{
  if (_iterator_on_modules && _moduleIterator == _modules.end())
  {
    _moduleIterator = _add_modules.begin();
    _iterator_on_modules = false;
  }
  if (!_iterator_on_modules && _moduleIterator == _add_modules.end())
    return NULL;

  return *(_moduleIterator++);
}

void Core::addService(const std::string& name, Service* service)
{
  pair<map<string, Service*>::iterator, bool> result;
  result = _services.insert(pair<string, Service*>(name, service));
  if (result.second == false)
    logFatal("Service already exists: %s", name.c_str());

  logInfo("Set service: %s", name.c_str());
}

Service* Core::removeService(const std::string& name)
{
  map<string, Service*>::iterator result = _services.find(name);
  if (result == _services.end())
    logFatal("Could not find service: %s", name.c_str());

  Service* service = result->second;
  _services.erase(result);
  logInfo("Removed service: %s", name.c_str());
  return service;
}

Service& Core::getService(const std::string& name) const
{
  map<string, Service*>::const_iterator result = _services.find(name);
  if (result == _services.end())
    logFatal("Could not find service: %s", name.c_str());

  return *result->second;
}

size_t Core::getServiceCount(void) const
{
  return _services.size();
}

void Core::addLog(const std::string& name, Log* log)
{
  pair<map<string, Log*>::iterator, bool> result;
  result = _logs.insert(pair<string, Log*>(name, log));
  if (result.second == false)
    logFatal("Log already exists: %s", name.c_str());

  logInfo("Set log: %s", name.c_str());
}

Log* Core::removeLog(const std::string& name)
{
  map<string, Log*>::iterator result = _logs.find(name);
  if (result == _logs.end())
    logFatal("Could not find log: %s", name.c_str());

  Log* log = result->second;
  _logs.erase(result);
  logInfo("Removed log: %s", name.c_str());
  return log;
}

Log& Core::getLog(const std::string& name) const
{
  map<string, Log*>::const_iterator result = _logs.find(name);
  if (result == _logs.end())
    logFatal("Could not find log: %s", name.c_str());

  return *result->second;
}

size_t Core::getLogCount(void) const
{
  return _logs.size();
}

void Core::run(void)
{
  /* Load modules, adding new dependencies if they are found. */
  while (!_add_modules.empty())
  {
    /*
     * Merge the current and newly added module lists. The newly added module
     * list is required because the main list may have active iterators on
     * it while an addModule() method call takes place.
     */
    size_t count = _modules.size();
    _modules.resize(count + _add_modules.size());
    copy(_add_modules.begin(), _add_modules.end(), _modules.begin() + count);
    _add_modules.clear();

    for_each(_modules.begin(), _modules.end(), mem_fun(&Module::_load));
  }

  if (_print_help)
  {
    _printHelp();
    return;
  }

  /* Check to make sure all options were valid. */
  for_each(_modules.begin(), _modules.end(), mem_fun(&Module::_checkOptions));

  if (_modules.empty())
    logFatal("No modules found");

  /* Start the modules. */
  for_each(_modules.begin(), _modules.end(), mem_fun(&Module::_start));

  logNotice("Core now running");

  while (!_shutdown_request)
  {
    /* See if any modules need to be run or reconfigured. */
    for_each(_modules.begin(), _modules.end(), mem_fun(&Module::_run));
    for_each(_modules.begin(), _modules.end(), mem_fun(&Module::_reconfigure));
    _run_notification.wait();
  }
}

void Core::shutdown(void)
{
  logNotice("Core received shutdown request");
  _shutdown_request = true;
  _run_notification.send();
}

/*
 * Private methods.
 */

const string& Core::_getAlias(const std::string& name) const
{
  for (map<string, string>::const_iterator alias = _aliases.begin();
       alias != _aliases.end();
       ++alias)
  {
    if (!strcasecmp(alias->first.c_str(), name.c_str()))
      return alias->second;
  }

  return name;
}

Module* Core::_getModule(const std::string& name) const
{
  /*
   * The find_if algorithm cannot be used here because this would require
   * bind2nd, and that has the reference-to-reference problem with some
   * implementations when used with methods that take a reference.
   */
  for (vector<Module*>::const_iterator module = _modules.begin();
       module != _modules.end();
       ++module)
  {
    if (!strcasecmp((*module)->getName().c_str(), name.c_str()))
      return *module;
  }

  for (vector<Module*>::const_iterator module = _add_modules.begin();
       module != _add_modules.end();
       ++module)
  {
    if (!strcasecmp((*module)->getName().c_str(), name.c_str()))
      return *module;
  }

  return NULL;
}

void Core::_setFatalMessage(const char* format, std::va_list arguments) const
{
  char message[SCALESTACK_MAX_MESSAGE_SIZE];
  vsnprintf(message, SCALESTACK_MAX_MESSAGE_SIZE, format, arguments);
  throw Common::Exception("Kernel", message);
}

void Core::_logMessage(Log::Level level, const char* format,
                       std::va_list arguments) const
{
  _log(level, "Kernel", format, arguments);
}

void Core::_log(Log::Level level, const char* context, const char* format,
                std::va_list arguments) const
{
  char message[SCALESTACK_MAX_MESSAGE_SIZE];
  vsnprintf(message, SCALESTACK_MAX_MESSAGE_SIZE, format, arguments);

  map<string, Log*>::const_iterator log;
  for (log = _logs.begin(); log != _logs.end(); ++log)
    log->second->log(level, context, message);

  if (level == Log::LEVEL_FATAL)
    throw Common::Exception(context, message);
}

void Core::_printHelp(void)
{
  cout << endl << PACKAGE_STRING << endl << "\n\
Kernel\n\
  v - Increase verbosity level.\n\
  h, help - Display this help message.\n\
  p, path=PATH - Add path to module search list.\n\
                 (Default: " << SCALESTACK_QUOTE_VALUE(SCALESTACK_MODULE_PATH) << ")\n\
\n";

  for_each(_modules.begin(), _modules.end(), mem_fun(&Module::_printHelp));

  cout << "\n\
Modules can be loaded by simply giving the module name as an\n\
option. For example, if you run 'scalestack Echo', this will load\n\
the Echo module and any dependencies.\n\
\n\
Module aliases may be specified using the format:\n\
\n\
  ModuleAlias=ModuleName\n\
\n\
For example, if you run 'scalestack EventService=EventLibevent', the\n\
module 'EventLibevent' will be loaded if any other modules depend\n\
on 'EventService'. This allows a generic service name to be listed\n\
as a dependency when there are multiple modules that could provide\n\
that service.\n\
\n\
Module options are specified using the format\n\
\n\
  ModuleName.option[=value]\n\
\n\
For example, if you run 'scalestack EchoServer.ports=4321', the\n\
EchoServer module will be loaded and will be run on port '4321'. For\n\
boolean options, you can specify 'false', 'f', 'no', 'n', or '0' to\n\
mean false using any case. Everything else will be treated as true.\n\
\n\
";
}

} /* namespace Kernel */

} /* namespace ScaleStack */
