/*
 * Scale Stack
 *
 * Copyright (C) 2010 Eric Day (eday@oddments.org)
 * All rights reserved.
 *
 * Use and distribution licensed under the BSD license. See the
 * COPYING file in the root project directory for full text.
 */

/**
 * @file
 * @brief Kernel Core Declarations
 */

#ifndef SCALESTACK_KERNEL_CORE_H
#define SCALESTACK_KERNEL_CORE_H

#include <map>
#include <string>
#include <vector>

#include <ScaleStack/Common/Macros.h>
#include <ScaleStack/Common/Notification.h>
#include <ScaleStack/Kernel/LogMessage.h>

namespace ScaleStack
{

namespace Kernel
{

class Module;
class Service;
class Log;

/**
 * This class is the central part of the kernel and is responsible for
 * managing modules, services, and kernel logs.
 */
class SCALESTACK_API Core: private LogMessage
{
  friend class Module;

public:

  /**
   * Constructor. This initializes all data members and adds a LogConsole
   * object to the list of logs.
   */
  Core(void);

  /**
   * Destructor. This ensures all modules are properly cleaned up and
   * generates error messages if there are leftover services or logs.
   */
  ~Core();

  /**
   * Parse arguments, most likely from the main() function arguments.
   *
   * @param[in] argument_count Argument count.
   * @param[in] arguments Array of argument values.
   */
  void parseArguments(int argument_count, const char** arguments);

  /**
   * Add an option value for a module. If a value already exists, the old
   * value is overwritten. Options can be of the form:
   *
   * module_name - Add the module.
   * module_name=value - Add a module alias from module_name to value.
   * module_name.option_name=value - Set a module option.
   *
   * @param[in] module_name Name of the module to add the option for.
   * @param[in] option_name Name of the option.
   * @param[in] value Value of the option.
   */
  void addModuleOptionValue(const std::string& module_name,
                            const std::string& option_name,
                            const std::string& value);

  /**
   * Set the log threshold for kernel logs. This is also set by passing 'v'
   * or 'verbose' arguments into parseArguments().
   *
   * @param[in] log_threshold The new log threshold to set.
   */
  void setLogThreshold(size_t log_threshold);

  /**
   * Set common signal handlers. Calling applications can either use this for
   * convenience or handle their own signals. SIGPIPE is set to be ignored,
   * and the following signals are set to call the given handler function:
   * SIGHUP, SIGINT, SIGQUIT, SIGTERM, SIGUSR1, and SIGUSR2.
   */
  void setSignalHandlers(void);

  /**
   * Handle signals received by application signal handlers. This currently
   * process SIGINT, SIGQUIT, and SIGTERM as shutdown requests. This can be
   * used by applications that wish to handle their own signals.
   *
   * @param[in] signal_number Signal that was caught and needs to be handled.
   */
  void handleSignal(int signal_number);

  /**
   * Get a module that has been added to the core, or add one if a module
   * by that name doesn't exist.
   *
   * @param[in] name Name of the module to add. This is used in constructing
   *  the path of the module to load. Name comparisons are case insensitive.
   * @return The module that was added.
   */
  Module& getOrAddModule(const std::string& name);

  /**
   * Get a module that has been added to the core.
   *
   * @param[in] name Name of the module to look for. This is used to check
   *  both the name and alias list. Name comparisons are case insensitive.
   * @return The module that was found.
   */
  Module& getModule(const std::string& name) const;

  /**
   * Get a count of the number of modules in the core.
   */
  size_t getModuleCount(void) const;

  /**
   * Reset the module iterator to the beginning and return first module. This
   * should be called first before iterating through all modules. The most
   * common way to use this interface is:
   *
   * @code
   *   while (Module* module = core.beginModule();
   *          module != NULL;
   *          module = core.nextModule())
   *   {
   *     // Do something with module here.
   *   }
   * @endcode
   *
   * @return First module or NULL if there are no modules.
   */
  Module* beginModule(void);

  /**
   * Get the next module while iterating, see beginModule() for usage details.
   *
   * @return Next module or NULL if at the end of the list.
   */
  Module* nextModule(void);

  /**
   * Add a service object for the given name. If a service by that name
   * already exists, this will fail and return false.
   *
   * @param[in] name Name of the service to add.
   * @param[in] service The service object to add.
   */
  void addService(const std::string& name, Service* service);

  /**
   * Remove a service object for the given name.
   *
   * @param[in] name Name of the service to remove.
   * @return The service object that was removed.
   */
  Service* removeService(const std::string& name);

  /**
   * Get a service object for the given name.
   *
   * @param[in] name Name of the service to get.
   * @return The service object that was requested.
   */
  Service& getService(const std::string& name) const;

  /**
   * Get a count of the number of services that have been added.
   */
  size_t getServiceCount(void) const;

  /**
   * Add a log object for the given name. If a log by that name already
   * exists, this will fail and return false.
   *
   * @param[in] name Name of the log to add.
   * @param[in] log The log object to add.
   */
  void addLog(const std::string& name, Log* log);

  /**
   * Remove a log object for the given name.
   *
   * @param[in] name Name of the log to remove.
   * @return The log object that was removed.
   */
  Log* removeLog(const std::string& name);

  /**
   * Get a log object for the given name.
   *
   * @param[in] name Name of the log to get.
   * @return The log object that was requested.
   */
  Log& getLog(const std::string& name) const;

  /**
   * Get a count of the number of logs that have been added.
   */
  size_t getLogCount(void) const;

  /**
   * This is the main loop of the core and will continue running until all
   * modules have been removed or a shutdown request is received.
   */
  void run(void);

  /**
   * Notify the kernel core that it should start the shutdown process. This is
   * usually called from signal handlers or modules providing administrative
   * functions.
   */
  void shutdown(void);

private:

  /**
   * Don't allow copying of objects.
   */
  SCALESTACK_LOCAL
  Core(const Core&);

  /**
   * Don't allow assignment of objects.
   */
  SCALESTACK_LOCAL
  Core& operator=(const Core&);

  /**
   * Resolve module names using the alias map.
   *
   * @param[in] name Name to look up in alias map.
   * @return Alias if found, or original name that was passed in.
   */
  SCALESTACK_LOCAL
  const std::string& _getAlias(const std::string& name) const;

  /**
   * Get a module that has been added to the core, this is only used
   * internally. This behaves just like getModule() (and is used by that
   * method) but does not throw exceptions when the module is not found.
   */
  SCALESTACK_LOCAL
  Module* _getModule(const std::string& name) const;

  /**
   * Set a fatal message as the error string. This is declared const so
   * other const methods can call it.
   *
   * @param[in] format Message format string.
   * @param[in] arguments Argument list for formatted message.
   */
  SCALESTACK_LOCAL
  void _setFatalMessage(const char* format, std::va_list arguments) const;

  /**
   * Log a core message. This is declared const so other const methods can
   * call it.
   *
   * @param[in] level Log level for this message.
   * @param[in] format Message format string.
   * @param[in] arguments Argument list for formatted message.
   */
  SCALESTACK_LOCAL
  void _logMessage(Log::Level level, const char* format,
                   std::va_list arguments) const;

  /**
   * Send a log message to all configured logs. This is declared const so
   * other const methods can call it.
   *
   * @param[in] level Log level for this message.
   * @param[in] context Context string for this message.
   * @param[in] format Message format string.
   * @param[in] arguments Argument list for formatted message.
   */
  SCALESTACK_LOCAL
  void _log(Log::Level level, const char* context, const char* format,
            std::va_list arguments) const;

  /**
   * Print a help message for the kernel and all loaded modules with options.
   */
  SCALESTACK_LOCAL
  void _printHelp(void);

  bool _print_help;
  bool _shutdown_request;
  bool _iterator_on_modules;
  Common::Notification _run_notification;
  std::vector<std::string> _paths;
  std::vector<Module*> _modules;
  std::vector<Module*> _add_modules;
  std::vector<Module*>::iterator _moduleIterator;
  std::map<std::string, std::string> _aliases;
  std::map<std::string, Service*> _services;
  std::map<std::string, Log*> _logs;
};

} /* namespace Kernel */

} /* namespace ScaleStack */

#endif /* SCALESTACK_KERNEL_CORE_H */
