/*
 * Scale Stack
 *
 * Copyright (C) 2010 Eric Day (eday@oddments.org)
 * All rights reserved.
 *
 * Use and distribution licensed under the BSD license. See the
 * COPYING file in the root project directory for full text.
 */

/**
 * @file
 * @brief Kernel Core Tests
 */

#include <ScaleStack/Common/Test.h>

#include <cstring>

#include <ScaleStack/Kernel/Service.h>

TEST_SUITE_BEGIN

TEST_BEGIN(constructor)
TEST_END

TEST_BEGIN(parseArguments)
  const char* argv[] =
  {
    "v",
    "-h",
    "--help",
    "TestChild"
    "TestChild=Value1",
    "TestParent.Option1",
    "TestParent.Option2=Value2"
  };
  core.parseArguments(sizeof(argv) / sizeof(const char*), argv);
TEST_END

TEST_BEGIN(parsePaths)
  const char* argv[] =
  {
    "p=/path/to/modules",
    "path=/another/path/to/modules"
  };
  core.parseArguments(sizeof(argv) / sizeof(const char*), argv);
TEST_END

TEST_BEGIN(parseBadPaths)
  const char* argv[] =
  {
    "p",
    "path"
  };
  TEST_EXCEPTION(core.parseArguments(sizeof(argv) / sizeof(const char*), argv));
TEST_END

TEST_BEGIN(badOption)
  const char* argv[] =
  {
    ""
  };
  TEST_EXCEPTION(core.parseArguments(sizeof(argv) / sizeof(const char*), argv));
  argv[0] = ".option=value";
  TEST_EXCEPTION(core.parseArguments(sizeof(argv) / sizeof(const char*), argv));
  argv[0] = "=value";
  TEST_EXCEPTION(core.parseArguments(sizeof(argv) / sizeof(const char*), argv));
TEST_END

TEST_BEGIN_NO_CORE(longOption)
  Kernel::Core core;
  char long_option[1024*128];
  const char* argv[] =
  {
    long_option
  };

  memset(long_option, 'A', sizeof(long_option) - 1);
  long_option[sizeof(long_option) - 1] = '\0';
  long_option[sizeof(long_option) / 3] = '.';
  long_option[(sizeof(long_option) / 3) * 2] = '=';

  core.parseArguments(sizeof(argv) / sizeof(const char*), argv);
TEST_END

TEST_BEGIN_NO_CORE(longOptionList)
  Kernel::Core core;
  const char* argv[1024*128];

  for (int x = 0; x < 1024*128; x++)
    argv[x] = const_cast<char*>("Module.Option=Value");

  core.parseArguments(sizeof(argv) / sizeof(const char*), argv);
TEST_END

TEST_BEGIN(addModuleOptionValue)
  core.addModuleOptionValue("Module", "Option", "Value");
  core.getModule("Module");
  TEST_EXCEPTION(core.addModuleOptionValue("", "Option", "Value"));
  core.addModuleOptionValue("Module", "", "Value");
  core.addModuleOptionValue("Module", "Option", "");
  TEST_EXCEPTION(core.addModuleOptionValue("", "", "Value"));
  TEST_EXCEPTION(core.addModuleOptionValue("", "Option", ""));
  core.addModuleOptionValue("Module", "", "");
  TEST_EXCEPTION(core.addModuleOptionValue("", "", ""));
TEST_END

TEST_BEGIN(setSignalHandlers)
  core.setSignalHandlers();
  TEST_EXCEPTION(core.setSignalHandlers());
TEST_END

TEST_BEGIN(handleSignal)
  core.handleSignal(0);
TEST_END

TEST_BEGIN(getOrAddModule)
  Kernel::Module& module = core.getOrAddModule("Module");
  assert(&core.getOrAddModule("Module") == &module);

  Kernel::Module& module2 = core.getOrAddModule("Module2");
  assert(&core.getOrAddModule("Module2") == &module2);
  assert(&core.getOrAddModule("Module2") != &module);
TEST_END

TEST_BEGIN(getModule)
  Kernel::Module& module = core.getOrAddModule("Module");
  assert(&core.getModule("Module") == &module);
  TEST_EXCEPTION(core.getModule("BadModule"));
TEST_END

TEST_BEGIN(getModuleCount)
  assert(core.getModuleCount() == 0);
  core.getOrAddModule("Module");
  assert(core.getModuleCount() == 1);
  core.getOrAddModule("Module2");
  assert(core.getModuleCount() == 2);
  core.getOrAddModule("Module2");
  assert(core.getModuleCount() == 2);
TEST_END

TEST_BEGIN(moduleScan)
  assert(core.beginModule() == NULL);
  assert(core.nextModule() == NULL);

  Kernel::Module& module = core.getOrAddModule("Module");
  assert(core.beginModule() == &module);
  assert(core.nextModule() == NULL);

  Kernel::Module& module2 = core.getOrAddModule("Module2");
  assert(core.beginModule() == &module);
  assert(core.nextModule() == &module2);
  assert(core.nextModule() == NULL);
TEST_END

TEST_BEGIN(addService)
  Kernel::Module& module = core.getOrAddModule("Module");
  Kernel::Service service(module);;
  core.addService("Service1", &service);
  core.addService("Service2", &service);
  TEST_EXCEPTION(core.addService("Service2", &service));
  assert(&service.getModule() == &module);
TEST_END

TEST_BEGIN(removeService)
  Kernel::Module& module = core.getOrAddModule("Module");
  Kernel::Service service(module);
  Kernel::Service service2(module);
  core.addService("Service1", &service);
  core.addService("Service2", &service2);
  assert(core.removeService("Service1") == &service);
  assert(core.removeService("Service2") == &service2);
  TEST_EXCEPTION(core.removeService("Service2"));
TEST_END

TEST_BEGIN(getService)
  Kernel::Module& module = core.getOrAddModule("Module");
  Kernel::Service service(module);
  Kernel::Service service2(module);
  core.addService("Service1", &service);
  core.addService("Service2", &service2);
  assert(&core.getService("Service1") == &service);
  assert(&core.getService("Service1") == &service);
  assert(&core.getService("Service2") == &service2);
  TEST_EXCEPTION(core.getService("Service3"));
TEST_END

class LogTest: public Kernel::Log
{
public:
  size_t _errors;
  LogTest(void): _errors() {}
  void log(Level level, const char*, const char*)
  {
    if (level == Kernel::Log::LEVEL_ERROR)
      _errors++;
  }
};

TEST_BEGIN_NO_CORE(addLog)
  Kernel::Core core;
  LogTest log;
  core.addLog("Log1", &log);
  core.addLog("Log2", &log);
  TEST_EXCEPTION(core.addLog("Log2", &log));
TEST_END

TEST_BEGIN_NO_CORE(removeLog)
  Kernel::Core core;
  LogTest log;
  LogTest log2;
  core.addLog("Log1", &log);
  core.addLog("Log2", &log2);
  assert(core.removeLog("Log1") == &log);
  assert(core.removeLog("Log2") == &log2);
  TEST_EXCEPTION(core.removeLog("Log2"));
TEST_END

TEST_BEGIN_NO_CORE(getLog)
  Kernel::Core core;
  LogTest log;
  LogTest log2;
  core.addLog("Log1", &log);
  core.addLog("Log2", &log2);
  assert(&core.getLog("Log1") == &log);
  assert(&core.getLog("Log1") == &log);
  assert(&core.getLog("Log2") == &log2);
  TEST_EXCEPTION(core.getLog("Log3"));
TEST_END

TEST_BEGIN_NO_CORE(unremovedLog)
  cout << "Testing unremoved logs, ignore the following output:" << endl;
  LogTest log;
  /* Put this in a nested scope so it can destruct before end of test. */
  {
    Kernel::Core core;

    /* Turn on verbose so we can catch errors. */
    const char* argv[] =
    {
      "v"
    };
    core.parseArguments(sizeof(argv) / sizeof(const char*), argv);

    /* Core destructor errors on more than one unremoved log because it
       assumes Console is there. Add two logs so more than one still exists. */
    core.addLog("Log1", &log);
    core.addLog("Log2", &log);

    /* Core goes out of scope here and destructs. */
  }

  /* There should be 6 errors, 3 for each set log (header and two log names). */
  assert(log._errors == 8);
TEST_END

TEST_BEGIN(run)
  cout << "Testing exception message, ignore the following output:" << endl;
  bool run_failed = false;
  try
  {
    core.run();
  }
  catch (exception& e)
  {
    cout << e.what() << endl;
    run_failed = true;
  }
  assert(run_failed);
TEST_END

TEST_BEGIN(runVerbose)
  const char* argv[] =
  {
    "vvv",
  };
  core.parseArguments(sizeof(argv) / sizeof(const char*), argv);
  cout << "Testing verbose messages, ignore the following output:" << endl;
  TEST_EXCEPTION(core.run());
TEST_END

TEST_BEGIN(shutdown)
  core.shutdown();
TEST_END

TEST_BEGIN(printHelp)
  const char* argv[] =
  {
    "help",
  };
  core.parseArguments(sizeof(argv) / sizeof(const char*), argv);
  cout << "Testing help messages, ignore the following output:" << endl;
  core.run();
TEST_END

TEST_BEGIN(alias)
  Kernel::Module& module = core.getOrAddModule("Module");
  Kernel::Module& module2 = core.getOrAddModule("Module2");
  core.addModuleOptionValue("ModuleA", "", "Module");
  assert(&core.getModule("ModuleA") == &module);
  core.addModuleOptionValue("ModuleA", "", "Module2");
  assert(&core.getModule("ModuleA") == &module2);
TEST_END

TEST_SUITE_END
