/*
 * Scale Stack
 *
 * Copyright (C) 2010 Eric Day (eday@oddments.org)
 * All rights reserved.
 *
 * Use and distribution licensed under the BSD license. See the
 * COPYING file in the root project directory for full text.
 */

/**
 * @file
 * @brief Log Declarations
 */

#ifndef SCALESTACK_KERNEL_LOG_H
#define SCALESTACK_KERNEL_LOG_H

#include <ScaleStack/Common/Macros.h>

namespace ScaleStack
{

namespace Kernel
{

/**
 * This is the base class for all kernel log implementations.
 */
class SCALESTACK_API Log
{
public:

  /**
   * Available log levels.
   */
  enum Level
  {
    LEVEL_FATAL,
    LEVEL_ERROR,
    LEVEL_NOTICE,
    LEVEL_INFO,
    LEVEL_DEBUG
  };

  Log(void);

  virtual ~Log();

  /**
   * Process a log entry, this must be supplied by derived classes.
   *
   * @param[in] level Log level for this message.
   * @param[in] context Context string for this message.
   * @param[in] message Message to log.
   */
  virtual void log(Level level, const char* context, const char* message) = 0;

  /**
   * Get the string representation for a givel log level.
   */
  const char* getLevelString(Level level);

private:

  /**
   * Don't allow copying of objects.
   */
  Log(const Log&);

  /**
   * Don't allow assignment of objects.
   */
  Log& operator=(const Log&);
};

/*
 * Public methods.
 */

inline Log::Log(void)
{
}

inline Log::~Log()
{
}

inline const char* Log::getLevelString(Level level)
{
  switch (level)
  {
  case LEVEL_FATAL:
    return "FATAL";
  case LEVEL_ERROR:
    return "ERROR";
  case LEVEL_NOTICE:
    return "NOTICE";
  case LEVEL_INFO:
    return "INFO";
  case LEVEL_DEBUG:
    return "DEBUG";
  default:
    return "(BAD)";
  };
}

} /* namespace Kernel */

} /* namespace ScaleStack */

#endif /* SCALESTACK_KERNEL_LOG_H */
