/*
 * Scale Stack
 *
 * Copyright (C) 2010 Eric Day (eday@oddments.org)
 * All rights reserved.
 *
 * Use and distribution licensed under the BSD license. See the
 * COPYING file in the root project directory for full text.
 */

/**
 * @file
 * @brief Kernel Log Message Base Class
 */

#ifndef SCALESTACK_KERNEL_LOGMESSAGE_H
#define SCALESTACK_KERNEL_LOGMESSAGE_H

#include <cstdarg>

#include <ScaleStack/Common/Exception.h>
#include <ScaleStack/Kernel/Log.h>

namespace ScaleStack
{

namespace Kernel
{

/**
 * Base class available for Kernel classes that produce log messages.
 */
class SCALESTACK_API LogMessage
{
public:

  LogMessage(size_t log_threshold);

  virtual ~LogMessage();

  /**
   * Log a fatal message with the formatted string if log threshold is met.
   *
   * @param[in] format Message format string.
   */
  void logFatal(const char* format, ...) const;

  /**
   * Log an error message with the formatted string if log threshold is met.
   *
   * @param[in] format Message format string.
   */
  void logError(const char* format, ...) const;

  /**
   * Log a notice message with the formatted string if log threshold is met.
   *
   * @param[in] format Message format string.
   */
  void logNotice(const char* format, ...) const;

  /**
   * Log an info message with the formatted string if log threshold is met.
   *
   * @param[in] format Message format string.
   */
  void logInfo(const char* format, ...) const;

  /**
   * Log a debug message with the formatted string if log threshold is met.
   *
   * @param[in] format Message format string.
   */
  void logDebug(const char* format, ...) const;

protected:

  size_t _log_threshold;

private:

  /**
   * Don't allow copying of objects.
   */
  LogMessage(const LogMessage&);

  /**
   * Don't allow assignment of objects.
   */
  LogMessage& operator=(const LogMessage&);

  /**
   * Set a fatal error message. This is used so fatal messages can be caught
   * for error strings even if logging threshold is not met.
   *
   * @param[in] format Message format string.
   * @param[in] arguments Argument list for formatted message.
   */
  virtual void _setFatalMessage(const char* format,
                                std::va_list arguments) const = 0;

  /**
   * Log the message.
   *
   * @param[in] level Log level for this message.
   * @param[in] format Message format string.
   * @param[in] arguments Argument list for formatted message.
   */
  virtual void _logMessage(Log::Level level, const char* format,
                           std::va_list arguments) const = 0;
};

/*
 * Public methods.
 */

inline LogMessage::LogMessage(size_t log_threshold):
  _log_threshold(log_threshold)
{
}

inline LogMessage::~LogMessage()
{
}

inline void LogMessage::logFatal(const char* format, ...) const
{
  std::va_list arguments;
  va_start(arguments, format);

  /*
   * We know this is going to trigger a catch because fatal messages throw
   * exceptions, so this is done so we can call va_end before passing on
   * the exception.
   */
  try
  {
    if (_log_threshold >= 1)
      _logMessage(Log::LEVEL_FATAL, format, arguments);
    else
      _setFatalMessage(format, arguments);
  }
  catch (Common::Exception&)
  {
    va_end(arguments);
    throw;
  }

  va_end(arguments);
}

inline void LogMessage::logError(const char* format, ...) const
{
  if (_log_threshold >= 1)
  {
    std::va_list arguments;
    va_start(arguments, format);
    _logMessage(Log::LEVEL_ERROR, format, arguments);
    va_end(arguments);
  }
}

inline void LogMessage::logNotice(const char* format, ...) const
{
  if (_log_threshold >= 1)
  {
    std::va_list arguments;
    va_start(arguments, format);
    _logMessage(Log::LEVEL_NOTICE, format, arguments);
    va_end(arguments);
  }
}

inline void LogMessage::logInfo(const char* format, ...) const
{
  if (_log_threshold >= 2)
  {
    std::va_list arguments;
    va_start(arguments, format);
    _logMessage(Log::LEVEL_INFO, format, arguments);
    va_end(arguments);
  }
}

inline void LogMessage::logDebug(const char* format, ...) const
{
  if (_log_threshold >= 3)
  {
    std::va_list arguments;
    va_start(arguments, format);
    _logMessage(Log::LEVEL_DEBUG, format, arguments);
    va_end(arguments);
  }
}

} /* namespace Kernel */

} /* namespace ScaleStack */

#endif /* SCALESTACK_KERNEL_LOGMESSAGE_H */
