/*
 * Scale Stack
 *
 * Copyright (C) 2010 Eric Day (eday@oddments.org)
 * All rights reserved.
 *
 * Use and distribution licensed under the BSD license. See the
 * COPYING file in the root project directory for full text.
 */

/**
 * @file
 * @brief Module Definitions
 */

#include "config.h"

#include <algorithm>
#include <cctype>
#include <cerrno>
#include <dlfcn.h>
#include <exception>
#include <functional>
#include <iostream>
#include <strings.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#include <ScaleStack/Common/FunctionObjects.h>
#include <ScaleStack/Kernel/Core.h>
#include <ScaleStack/Kernel/Module.h>
#include <ScaleStack/Kernel/Option.h>

using namespace std;

namespace ScaleStack
{

namespace Kernel
{

/*
 * Public methods.
 */

Module::~Module()
{
  logInfo("Destructing");

  _unload();

  for_each(_options.begin(), _options.end(), Common::DeletePointer());

  logInfo("Destructed");
}

Core& Module::getCore(void) const
{
  return _core;
}

const string& Module::getName(void) const
{
  return _name;
}

const string& Module::getPathname(void) const
{
  return _pathname;
}

const string& Module::getVersion(void) const
{
  return _version;
}

const string& Module::getAuthor(void) const
{
  return _author;
}

const string& Module::getTitle(void) const
{
  return _title;
}

const string& Module::getLicense(void) const
{
  return _license;
}

const string& Module::getDependencies(void) const
{
  return _dependencies;
}

size_t Module::getParentCount(void) const
{
  return _parents.size();
}

Module* Module::beginParent(void)
{
  _parentIterator = _parents.begin();
  return nextParent();
}

Module* Module::nextParent(void)
{
  if (_parentIterator == _parents.end())
    return NULL;

  return *(_parentIterator++);
}

size_t Module::getChildCount(void) const
{
  return _children.size();
}

Module* Module::beginChild(void)
{
  _childIterator = _children.begin();
  return nextChild();
}

Module* Module::nextChild(void)
{
  if (_childIterator == _children.end())
    return NULL;

  return *(_childIterator++);
}

bool Module::isLoaded(void) const
{
  return _handle != NULL;
}

Option& Module::addOption(const std::string& name,
                          const std::string& help,
                          const std::string& value_name,
                          const std::string& default_value)
{
  Option& option = _getOrAddOption(name);
  if (option.isValid())
    logFatal("Duplicate option added: %s", name.c_str());

  option._setHelp(help);
  option._setValueName(value_name);
  option._setDefaultValue(default_value);
  option._setValue(default_value, false);
  option._setValid(true);
  return option;
}

Option& Module::getOption(const std::string& name) const
{
  Option* option = _getOption(name);
  if (option == NULL)
    logFatal("Could not find option: %s", name.c_str());

  return *option;
}

size_t Module::getOptionCount(void) const
{
  return _options.size();
}

Option* Module::beginOption(void)
{
  _optionIterator = _options.begin();
  return nextOption();
}

Option* Module::nextOption(void)
{
  if (_optionIterator == _options.end())
    return NULL;

  return *(_optionIterator++);
}

/*
 * Private methods.
 */

Module::Module(Core& core, const std::string& name, size_t log_threshold):
  LogMessage(log_threshold),
  _is_loading(false),
  _is_started(false),
  _need_run(true),
  _need_reconfigure(true),
  _core(core),
  _handle(),
  _definition(),
  _name(name),
  _pathname(),
  _version(),
  _author(),
  _title(),
  _license(),
  _dependencies(),
  _parents(),
  _parentIterator(_parents.begin()),
  _children(),
  _childIterator(_children.begin()),
  _options(),
  _optionIterator(_options.begin())
{
  logInfo("Constructed");
}

void Module::_load(void)
{
  if (isLoaded())
    return;

  if (_is_loading)
    logFatal("Module dependency loop detected");

  _loadDependencies();

  if (_dependencies.size() > 0)
  {
    size_t comma;
    size_t start = 0;

    do
    {
      comma = _dependencies.find(",", start);
      string module_name(_dependencies, start, comma - start);

      if (module_name.size() > 0)
      {
        Module& parent = _core.getOrAddModule(module_name);
        _parents.push_back(&parent);
        parent._children.push_back(this);
        parent._load();
      }

      start = comma + 1;
    }
    while (comma != string::npos);
  }

  _loadDefinition();
}

void Module::_loadDependencies(void)
{
  string filename;
  for (string::iterator c = _name.begin(); c != _name.end(); ++c)
  {
    if (*c == ':')
    {
      ++c;
      filename.push_back('_');
    }
    else
      filename.push_back(static_cast<char>(tolower(*c)));
  }

  vector<string>::const_iterator path;
  for (path = _core._paths.begin(); path != _core._paths.end(); ++path)
  {
    _pathname = *(path) + "/lib" + filename + "_module." + SCALESTACK_MODULE_EXT;
    logInfo("Checking pathname: %s", _pathname.c_str());

    struct stat stat_buffer;
    int stat_return = stat(_pathname.c_str(), &stat_buffer);
    if (stat_return == 0)
      break;
    else if (stat_return == -1 && errno != ENOENT)
      logFatal("Could not stat() pathname: %s (%d)", _pathname.c_str(), errno);
  }

  if (path == _core._paths.end())
    logFatal("Could not find");

  logInfo("Loading from: %s", _pathname.c_str());

  _openObject(RTLD_LAZY);

  _dependencies = _definition->dependencies;

  logDebug("Dependencies: %s", _dependencies.c_str());

  _closeObject();
  _is_loading = true;
}

void Module::_loadDefinition(void)
{
  _openObject(RTLD_NOW | RTLD_GLOBAL);

  _name = _definition->name;
  _version = _definition->version;
  _author = _definition->author;
  _title = _definition->title;
  _license = _definition->license;

  if (_definition->options != NULL)
    _definition->options(*this);

  _is_loading = false;

  logNotice("Loaded");
  logDebug("Version: %s", _version.c_str());
  logDebug("Author: %s", _author.c_str());
  logDebug("Title: %s", _title.c_str());
  logDebug("License: %s", _license.c_str());
}

void Module::_unload(void)
{
  if (!isLoaded())
    return;

  logInfo("Unloading");

  for_each(_children.begin(), _children.end(), mem_fun(&Module::_unload));

  logInfo("Stopping");

  if (_is_started && _definition->stop != NULL)
    _definition->stop(*this);

  _is_started = false;

  logNotice("Unloaded");

  _closeObject();
}

void Module::_start(void)
{
  if (_is_started)
    return;

  for_each(_parents.begin(), _parents.end(), mem_fun(&Module::_start));

  logInfo("Starting");

  if (_definition->start != NULL)
    _definition->start(*this);

  _is_started = true;
}

void Module::_run(void)
{
  if (!_need_run || _definition->run == NULL)
    return;

  logDebug("Running");
  _need_run = false;
  _definition->run(*this);
}

void Module::_reconfigure(void)
{
  if (!_need_reconfigure || _definition->reconfigure == NULL)
    return;

  logDebug("Reconfiguring");
  _need_reconfigure = false;
  _definition->reconfigure(*this);
}

void Module::_openObject(int flags)
{
  _handle = dlopen(_pathname.c_str(), flags);
  if (_handle == NULL)
    logFatal("dlopen failed: %s", dlerror());

  _definition = static_cast<Module::Definition*>(dlsym(_handle, "_scalestack_kernel_module_definition"));
  if (_definition == NULL)
  {
    /* Copy dlerror() into buffer because the dlclose() may clear the error. */
    char error[SCALESTACK_MAX_MESSAGE_SIZE];
    snprintf(error, SCALESTACK_MAX_MESSAGE_SIZE, "%s", dlerror());
    (void) dlclose(_handle);
    _handle = NULL;
    logFatal("dlsym failed: %s", error);
  }
}

void Module::_closeObject(void)
{
  if (dlclose(_handle) != 0)
    logFatal("dlclose failed: %s", dlerror());

  _definition = NULL;
  _handle = NULL;
}

Option* Module::_getOption(const std::string& name) const
{
  /*
   * The find_if algorithm cannot be used here because this would require
   * bind2nd, and that has the reference-to-reference problem with some
   * implementations when used with methods that take a reference.
   */
  for (vector<Option*>::const_iterator option = _options.begin();
       option != _options.end();
       ++option)
  {
    if (!strcasecmp((*option)->getName().c_str(), name.c_str()))
      return *option;
  }

  return NULL;
}

Option& Module::_getOrAddOption(const std::string& name)
{
  Option* option = _getOption(name);
  if (option != NULL)
    return *option;

  option = new Option(*this, name);
  try
  {
    _options.push_back(option);
  }
  catch (exception&)
  {
    delete option;
    throw;
  }

  return *option;
}

void Module::_setOptionValue(const std::string& name, const std::string& value)
{
  Option& option = _getOrAddOption(name);
  option._setValue(value, true);
}

void Module::_checkOptions(void) const
{
  vector<Option*>::const_iterator option = find_if(_options.begin(),
                                                   _options.end(),
                                                   not1(mem_fun(&Option::isValid)));
  if (option != _options.end())
    logFatal("Invalid option: %s", (*option)->getName().c_str());
}

void Module::_setFatalMessage(const char* format, std::va_list arguments) const
{
  char message[SCALESTACK_MAX_MESSAGE_SIZE];
  vsnprintf(message, SCALESTACK_MAX_MESSAGE_SIZE, format, arguments);
  throw Common::Exception(_name.c_str(), message);
}

void Module::_logMessage(Log::Level level, const char* format,
                         std::va_list arguments) const
{
  _core._log(level, _name.c_str(), format, arguments);
}

void Module::_printHelp(void)
{
  if (_options.empty())
    return;

  cout << _name << endl;
  for_each(_options.begin(), _options.end(), mem_fun(&Option::_printHelp));
  cout << endl;
}

} /* namespace Kernel */

} /* namespace ScaleStack */
