/*
 * Scale Stack
 *
 * Copyright (C) 2010 Eric Day (eday@oddments.org)
 * All rights reserved.
 *
 * Use and distribution licensed under the BSD license. See the
 * COPYING file in the root project directory for full text.
 */

/**
 * @file
 * @brief Kernel Module Declarations
 */

#ifndef SCALESTACK_KERNEL_MODULE_H
#define SCALESTACK_KERNEL_MODULE_H

#include <stdint.h>
#include <string>
#include <vector>

#include <ScaleStack/Common/Macros.h>
#include <ScaleStack/Kernel/LogMessage.h>

namespace ScaleStack
{

namespace Kernel
{

class Core;
class Option;

/**
 * This class represents a module (shared object) in the system. It is
 * responsible for loading, running, unloading, and managing options.
 */
class SCALESTACK_API Module: public LogMessage
{
  friend class Core;

public:

  /**
   * Struct that every module needs to expose globally so that the module
   * loader can find and load it. Modules should use the macros at the end
   * of this header file to define this struct.
   */
  typedef struct
  {
    uint32_t api_version;
    const char* name;
    const char* version;
    const char* author;
    const char* title;
    const char* license;
    const char* dependencies;
    void (*options)(Module& module);
    void (*start)(Module& module);
    void (*stop)(Module& module);
    void (*run)(Module& module);
    void (*reconfigure)(Module& module);
  } Definition;

  ~Module();

  /**
   * Get the Core object that this module belongs to.
   */
  Core& getCore(void) const;

  /**
   * Get the name of the module used for loading.
   */
  const std::string& getName(void) const;

  /**
   * Get the full pathname of the module used for loading.
   */
  const std::string& getPathname(void) const;

  /**
   * Get the version of the module.
   */
  const std::string& getVersion(void) const;

  /**
   * Get the author of the module.
   */
  const std::string& getAuthor(void) const;

  /**
   * Get the title of the module.
   */
  const std::string& getTitle(void) const;

  /**
   * Get the license of the module.
   */
  const std::string& getLicense(void) const;

  /**
   * Get the list of module dependencies that must be loaded before this
   * module.
   */
  const std::string& getDependencies(void) const;

  /**
   * Get a count of the modules in parent dependency list.
   */
  size_t getParentCount(void) const;

  /**
   * This behaves just like Core::beginModule(), but for parent dependencies
   * of this module.
   */
  Module* beginParent(void);

  /**
   * This behaves just like Core::nextModule(), but for parent dependencies
   * of this module.
   */
  Module* nextParent(void);

  /**
   * Get a count of the modules in child dependency list.
   */
  size_t getChildCount(void) const;

  /**
   * This behaves just like Core::beginModule(), but for child dependencies
   * of this module.
   */
  Module* beginChild(void);

  /**
   * This behaves just like Core::nextModule(), but for child dependencies
   * of this module.
   */
  Module* nextChild(void);

  /**
   * Check if the module is currently loaded.
   */
  bool isLoaded(void) const;

  /**
   * Add an option to a module. This is usually only called in the "options"
   * callback function specified in the module definition. If an option
   * already exists, NULL is returned and an error is set in the module
   * to notify that a duplicate error message should be returned once the
   * "options" callback has completed. This prevents module writers from
   * having to check the return code of every addOption() call.
   *
   * Options can be added after the "options" callback, but they will not
   * show up in the help menu, will not be allowed from the command line,
   * and error checking must be done in the module.
   *
   * When displaying help on the command line, the format will be:
   *
   * --ModuleName.name=value_name - help (Default: default_value)
   *
   * @param[in] name Name of the option.
   * @param[in] help Help string for of the option.
   * @param[in] value_name The name of the value to use in the help.
   * @param[in] default_value The default value to use for the option.
   * @return Option that was added.
   */
  Option& addOption(const std::string& name,
                    const std::string& help,
                    const std::string& value_name,
                    const std::string& default_value);

  /**
   * Get an option that has been added to the module.
   *
   * @param[in] name Name of the option to look for.
   * @return The option that was requested.
   */
  Option& getOption(const std::string& name) const;

  /**
   * Get a count of the number of options in the module.
   */
  size_t getOptionCount(void) const;

  /**
   * This behaves just like Core::beginModule(), but for options of this
   * module.
   */
  Option* beginOption(void);

  /**
   * This behaves just like Core::nextModule(), but for options of this
   * module.
   */
  Option* nextOption(void);

private:

  /**
   * Construct a module instance. This is private because only Core, which
   * is listed as a friend, can construct modules.
   *
   * @param[in] core Core that created this module.
   * @param[in] name Name of module to create.
   * @param[in] log_threshold Logging threshold.
   */
  SCALESTACK_LOCAL
  Module(Core& core, const std::string& name, size_t log_threshold);

  /**
   * Don't allow copying of objects.
   */
  SCALESTACK_LOCAL
  Module(const Module&);

  /**
   * Don't allow assignment of objects.
   */
  SCALESTACK_LOCAL
  Module& operator=(const Module&);

  /**
   * Load a module by first detecting it's dependencies, loading those
   * modules, and then loading the requested module. This also includes
   * calling the option callback in the module once loaded to get option
   * information.
   */
  SCALESTACK_LOCAL
  void _load(void);

  /**
   * Try loading the dependency list from the module based on module path
   * list and module name.
   */
  SCALESTACK_LOCAL
  void _loadDependencies(void);

  /**
   * Try loading the module definition structure using the pathname generated
   * in _loadDependencies.
   */
  SCALESTACK_LOCAL
  void _loadDefinition(void);

  /**
   * Unload the module, unloading child dependencies if needed in the process.
   */
  SCALESTACK_LOCAL
  void _unload(void);

  /**
   * Call the module start function, starting all parent dependencies
   * if needed.
   */
  SCALESTACK_LOCAL
  void _start(void);

  /**
   * Call the module run function if needed.
   */
  SCALESTACK_LOCAL
  void _run(void);

  /**
   * Call the module reconfigure function if needed.
   */
  SCALESTACK_LOCAL
  void _reconfigure(void);

  /**
   * Open the shared object in pathname and map module definition symbol.
   *
   * @param[in] flags Flags to pass to dlopen().
   */
  SCALESTACK_LOCAL
  void _openObject(int flags);

  /**
   * Close the shared object associated with module.
   */
  SCALESTACK_LOCAL
  void _closeObject(void);

  /**
   * Get an option that has been added to the core, this is only used
   * internally. This behaves just like getOption() (and is used by that
   * method) but does not throw exceptions when the option is not found.
   */
  SCALESTACK_LOCAL
  Option* _getOption(const std::string& name) const;

  /**
   * Get an option that has been added to the module, or add one if an option
   * by that name doesn't exist. Used internally by other methods.
   *
   * @param name Name of the option to add.
   * @return The option that was requested.
   */
  SCALESTACK_LOCAL
  Option& _getOrAddOption(const std::string& name);

  /**
   * Set an option value. If a value already exists, the old value is
   * overwritten.
   *
   * @param[in] name Name of the option.
   * @param[in] value Value of the option.
   */
  SCALESTACK_LOCAL
  void _setOptionValue(const std::string& name, const std::string& value);

  /**
   * Check to see if all options set for the module are valid.
   */
  SCALESTACK_LOCAL
  void _checkOptions(void) const;

  /**
   * Set a fatal message as the error string.
   *
   * @param[in] format Message format string.
   * @param[in] arguments Argument list for formatted message.
   */
  SCALESTACK_LOCAL
  void _setFatalMessage(const char* format, std::va_list arguments) const;

  /**
   * Log a module message.
   *
   * @param[in] level Log level for this message.
   * @param[in] format Message format string.
   * @param[in] arguments Argument list for formatted message.
   */
  SCALESTACK_LOCAL
  void _logMessage(Log::Level level, const char* format,
                   std::va_list arguments) const;

  /**
   * Print a help message for this module.
   */
  SCALESTACK_LOCAL
  void _printHelp(void);

  bool _is_loading;
  bool _is_started;
  bool _need_run;
  bool _need_reconfigure;
  Core& _core;
  void* _handle;
  Definition* _definition;
  std::string _name;
  std::string _pathname;
  std::string _version;
  std::string _author;
  std::string _title;
  std::string _license;
  std::string _dependencies;
  std::vector<Module*> _parents;
  std::vector<Module*>::iterator _parentIterator;
  std::vector<Module*> _children;
  std::vector<Module*>::iterator _childIterator;
  std::vector<Option*> _options;
  std::vector<Option*>::iterator _optionIterator;
};

} /* namespace Kernel */

} /* namespace ScaleStack */

/**
 * Macro for creating a module definition structure. This uses information
 * from the plugin.ini file.
 */
#define SCALESTACK_KERNEL_MODULE(options, start, stop, run, reconfigure) \
  SCALESTACK_API ScaleStack::Kernel::Module::Definition _scalestack_kernel_module_definition = \
  { \
    SCALESTACK_CURRENT_API_VERSION, \
    SCALESTACK_QUOTE_VALUE(PANDORA_MODULE_NAME), \
    PANDORA_MODULE_VERSION, \
    PANDORA_MODULE_AUTHOR, \
    PANDORA_MODULE_TITLE, \
    SCALESTACK_QUOTE_VALUE(PANDORA_MODULE_LICENSE), \
    PANDORA_MODULE_DEPENDENCIES, \
    options, \
    start, \
    stop, \
    run, \
    reconfigure \
  }

#endif /* SCALESTACK_KERNEL_MODULE_H */
