/*
 * Scale Stack
 *
 * Copyright (C) 2010 Eric Day (eday@oddments.org)
 * All rights reserved.
 *
 * Use and distribution licensed under the BSD license. See the
 * COPYING file in the root project directory for full text.
 */

/**
 * @file
 * @brief Kernel::Module Tests
 */

#include <ScaleStack/Common/Test.h>

#include <signal.h>
#include <sys/types.h>
#include <unistd.h>

#include <ScaleStack/Kernel/Module.h>

TEST_SUITE_BEGIN

TEST_BEGIN(constructor)
  Kernel::Module& module = core.getOrAddModule("Module");
  assert(&module.getCore() == &core);
  assert(module.getName() == "Module");
  assert(module.getPathname() == "");
  assert(module.getVersion() == "");
  assert(module.getAuthor() == "");
  assert(module.getTitle() == "");
  assert(module.getLicense() == "");
  assert(module.getDependencies() == "");
  assert(module.getParentCount() == 0);
  assert(module.getChildCount() == 0);
  assert(module.isLoaded() == false);
TEST_END

TEST_BEGIN(parentScan)
  Kernel::Module& module = core.getOrAddModule("Module");
  assert(module.beginParent() == NULL);
  assert(module.nextParent() == NULL);
TEST_END

TEST_BEGIN(childScan)
  Kernel::Module& module = core.getOrAddModule("Module");
  assert(module.beginChild() == NULL);
  assert(module.nextChild() == NULL);
TEST_END

TEST_BEGIN(addOption)
  Kernel::Module& module = core.getOrAddModule("Module");
  module.addOption("Option", "Help", "ValueName", "DefaultValue");
TEST_END

TEST_BEGIN(getOption)
  Kernel::Module& module = core.getOrAddModule("Module");
  Kernel::Option& option = module.addOption("Option", "Help", "ValueName", "DefaultValue");
  assert(&module.getOption("Option") == &option);
  TEST_EXCEPTION(module.getOption("BadOption"));
TEST_END

TEST_BEGIN(getOptionCount)
  Kernel::Module& module = core.getOrAddModule("Module");
  assert(module.getOptionCount() == 0);
  module.addOption("Option", "Help", "ValueName", "DefaultValue");
  assert(module.getOptionCount() == 1);
  module.addOption("Option2", "Help", "ValueName", "DefaultValue");
  assert(module.getOptionCount() == 2);
TEST_END

TEST_BEGIN(optionScan)
  Kernel::Module& module = core.getOrAddModule("Module");
  assert(module.beginOption() == NULL);
  assert(module.nextOption() == NULL);

  Kernel::Option& option = module.addOption("Option", "Help", "ValueName", "DefaultValue");
  assert(module.beginOption() == &option);
  assert(module.nextOption() == NULL);

  Kernel::Option& option2 = module.addOption("Option2", "Help", "ValueName", "DefaultValue");
  assert(module.beginOption() == &option);
  assert(module.nextOption() == &option2);
  assert(module.nextOption() == NULL);
TEST_END

TEST_BEGIN(loadGrandParent)
  Kernel::Module& module = core.getOrAddModule("Kernel::TestModuleGrandParent");
  assert(module.getTitle() == "");
  assert(module.isLoaded() == false);
  core.run();
  assert(module.getTitle() == "TestModuleGrandParent Title");
  assert(module.getParentCount() == 0);
  assert(module.getChildCount() == 0);
TEST_END

TEST_BEGIN(loadParent)
  Kernel::Module& module = core.getOrAddModule("Kernel::TestModuleParent");
  assert(module.getTitle() == "");
  assert(module.isLoaded() == false);
  TEST_EXCEPTION(core.getModule("Kernel::TestModuleChild"));
  core.run();

  Kernel::Module& grand_parent = core.getModule("Kernel::TestModuleGrandParent");

  assert(grand_parent.getTitle() == "TestModuleGrandParent Title");
  assert(grand_parent.getParentCount() == 0);
  assert(grand_parent.getChildCount() == 1);
  assert(grand_parent.beginChild() == &module);
  assert(grand_parent.nextChild() == NULL);

  assert(module.getTitle() == "TestModuleParent Title");
  assert(module.getParentCount() == 1);
  assert(module.beginParent() == &grand_parent);
  assert(module.nextParent() == NULL);
  assert(module.getChildCount() == 0);
TEST_END

TEST_BEGIN(loadChild)
  Kernel::Module& module = core.getOrAddModule("Kernel::TestModuleChild");
  assert(module.getTitle() == "");
  assert(module.isLoaded() == false);
  core.run();

  Kernel::Module& parent = core.getModule("Kernel::TestModuleParent");
  Kernel::Module& grand_parent = core.getModule("Kernel::TestModuleGrandParent");

  assert(parent.getTitle() == "TestModuleParent Title");
  assert(parent.getParentCount() == 1);
  assert(parent.beginParent() == &grand_parent);
  assert(parent.nextParent() == NULL);
  assert(parent.getChildCount() == 1);
  assert(parent.beginChild() == &module);
  assert(parent.nextChild() == NULL);

  assert(grand_parent.getTitle() == "TestModuleGrandParent Title");
  assert(grand_parent.getParentCount() == 0);
  assert(grand_parent.getChildCount() == 1);
  assert(grand_parent.beginChild() == &parent);
  assert(grand_parent.nextChild() == NULL);

  assert(module.getTitle() == "TestModuleChild Title");
  assert(module.getParentCount() == 1);
  assert(module.beginParent() == &parent);
  assert(module.nextParent() == NULL);
  assert(module.getChildCount() == 0);
TEST_END

TEST_BEGIN(openError)
  core.getOrAddModule("Kernel::TestModuleOpenError");
  TEST_EXCEPTION(core.run());
TEST_END

TEST_BEGIN(symError)
  core.getOrAddModule("Kernel::TestModuleSymError");
  TEST_EXCEPTION(core.run());
TEST_END

TEST_BEGIN(loadLoop)
  core.getOrAddModule("Kernel::TestModuleLoop");
  TEST_EXCEPTION(core.run());
TEST_END

TEST_BEGIN(moduleScan)
  Kernel::Module& module = core.getOrAddModule("Kernel::TestModuleGrandParent");
  assert(core.beginModule() == &module);
  assert(core.nextModule() == NULL);

  core.run();

  Kernel::Module& module2 = core.getOrAddModule("Kernel::TestModuleParent");

  assert(core.beginModule() == &module);
  assert(core.nextModule() == &module2);
  assert(core.nextModule() == NULL);
TEST_END

TEST_BEGIN_NO_CORE(verbose)
  cout << "Testing verbose messages, ignore the following output:" << endl;
  /* Put this in a nested scope so it can destruct before end of test. */
  {
    Kernel::Core core;
    core.addModuleOptionValue("path", "", "ScaleStack/.libs");
    const char* argv[] =
    {
      "vvv"
    };
    core.parseArguments(sizeof(argv) / sizeof(const char*), argv);
    core.getOrAddModule("Kernel::TestModuleGrandParent");
    core.run();
  }
TEST_END

TEST_BEGIN(signalShutdown)
  core.getOrAddModule("Kernel::TestModuleGrandParent");
  core.setSignalHandlers();
  kill(getpid(), SIGQUIT);
  core.run();
TEST_END

TEST_BEGIN(printHelp)
  core.getOrAddModule("Kernel::TestOption");
  core.getOrAddModule("Kernel::TestModuleGrandParent");
  const char* argv[] =
  {
    "help",
  };
  core.parseArguments(sizeof(argv) / sizeof(const char*), argv);
  cout << "Testing help messages, ignore the following output:" << endl;
  core.run();
TEST_END

TEST_BEGIN(checkInvalidOptions)
  core.getOrAddModule("Kernel::TestOption");
  core.addModuleOptionValue("Kernel::TestOption", "Invalid", "Value");
  TEST_EXCEPTION(core.run());
TEST_END

TEST_SUITE_END
