/*
 * Scale Stack
 *
 * Copyright (C) 2010 Eric Day (eday@oddments.org)
 * All rights reserved.
 *
 * Use and distribution licensed under the BSD license. See the
 * COPYING file in the root project directory for full text.
 */

/**
 * @file
 * @brief Option Definitions
 */

#include "config.h"

#include <cstdlib>
#include <iostream>
#include <strings.h>

#include <ScaleStack/Kernel/Option.h>

using namespace std;

namespace ScaleStack
{

namespace Kernel
{

/*
 * Public methods.
 */

Option::~Option()
{
}

Module& Option::getModule(void) const
{
  return _module;
}

const string& Option::getName(void) const
{
  return _name;
}

const string& Option::getHelp(void) const
{
  return _help;
}

const string& Option::getValueName(void) const
{
  return _value_name;
}

const string& Option::getDefaultValue(void) const
{
  return _default_value;
}

const string& Option::getValue(void) const
{
  return _value;
}

int Option::getIntValue(void) const
{
  return atoi(_value.c_str());
}

size_t Option::getSizeValue(void) const
{
  return strtoul(_value.c_str(), NULL, 0);
}

bool Option::getBoolValue(void) const
{
  if (!strcasecmp(_value.c_str(), "false") ||
      !strcasecmp(_value.c_str(), "f") ||
      !strcasecmp(_value.c_str(), "no") ||
      !strcasecmp(_value.c_str(), "n") ||
      _value == "0")
  {
    return false;
  }

  return true;
}

vector<string> Option::getListValue(void) const
{
  vector<string> list;
  string::size_type current = 0;

  do
  {
    string::size_type next = _value.find_first_of(",", current);
    list.push_back(_value.substr(current, next - current));
    current = _value.find_first_not_of(",", next);
  }
  while (current != string::npos);

  return list;
}

vector<pair<string,string> > Option::getHostListValue(void) const
{
  vector<string> list = getListValue();
  vector<pair<string,string> > host_list;

  for (vector<string>::const_iterator host = list.begin();
       host != list.end();
       ++host)
  {
    string::size_type port = host->find_first_of(":");
    if (port == string::npos)
      host_list.push_back(make_pair(*host, ""));
    else
    {
      host_list.push_back(make_pair(host->substr(0, port),
                                    host->substr(port + 1)));
    }
  }

  return host_list;
}

bool Option::isValid(void) const
{
  return _is_valid;
}

/*
 * Private methods.
 */

Option::Option(Module& module, const std::string& name):
  _is_valid(false),
  _is_set(false),
  _module(module),
  _name(name),
  _help(),
  _value_name(),
  _default_value(),
  _value()
{
}

void Option::_setHelp(const std::string& help)
{
  _help = help;
}

void Option::_setValueName(const std::string& value_name)
{
  _value_name = value_name;
}

void Option::_setDefaultValue(const std::string& default_value)
{
  _default_value = default_value;
}

void Option::_setValue(const std::string& value, bool overwrite)
{
  if (overwrite || !_is_set)
  {
    _value = value;
    _is_set = true;
  }
}

void Option::_setValid(bool is_valid)
{
  _is_valid = is_valid;
}

void Option::_printHelp(void)
{
  cout << "  " << _name << "=" << _value_name << " - " << _help;

  if (_default_value != "")
    cout << " (Default: " << _default_value << ")";

  cout << endl;
}

} /* namespace Kernel */

} /* namespace ScaleStack */
