/*
 * Scale Stack
 *
 * Copyright (C) 2010 Eric Day (eday@oddments.org)
 * All rights reserved.
 *
 * Use and distribution licensed under the BSD license. See the
 * COPYING file in the root project directory for full text.
 */

/**
 * @file
 * @brief Option Declarations
 */

#ifndef SCALESTACK_KERNEL_OPTION_H
#define SCALESTACK_KERNEL_OPTION_H

#include <string>
#include <utility>
#include <vector>

#include <ScaleStack/Common/Macros.h>

namespace ScaleStack
{

namespace Kernel
{

class Module;

/**
 * This class represents a module option and provides various parsing methods
 * for values.
 */
class SCALESTACK_API Option
{
  friend class Module;

public:

  ~Option();

  /**
   * Get the Module object this option belongs to.
   */
  Module& getModule(void) const;

  /**
   * Get the name of the option.
   */
  const std::string& getName(void) const;

  /**
   * Get the help message of the option.
   */
  const std::string& getHelp(void) const;

  /**
   * Get the value name of the option.
   */
  const std::string& getValueName(void) const;

  /**
   * Get the default value of the option.
   */
  const std::string& getDefaultValue(void) const;

  /**
   * Get the value of the option.
   */
  const std::string& getValue(void) const;

  /**
   * Get the value of the option converted to a int.
   */
  int getIntValue(void) const;

  /**
   * Get the value of the option converted to a size_t.
   */
  size_t getSizeValue(void) const;

  /**
   * Get the value of the option converted to a boolean.
   */
  bool getBoolValue(void) const;

  /**
   * Get the value of the option converted to a list. The format of the list
   * is comma seperated.
   */
  std::vector<std::string> getListValue(void) const;

  /**
   * Get the value of the option converted to a host list. The format of
   * the list is comma seperated host[:port] pairs.
   */
  std::vector<std::pair<std::string,std::string> > getHostListValue(void) const;

  /**
   * Check to see if the associated module asked for this to be a valid
   * option.
   */
  bool isValid(void) const;

private:

  /**
   * Construct an option instance.
   *
   * @param[in] module Module that created this module.
   * @param[in] name Name of option to create.
   */
  SCALESTACK_LOCAL
  Option(Module& module, const std::string& name);

  /**
   * Don't allow copying of objects.
   */
  SCALESTACK_LOCAL
  Option(const Option&);

  /**
   * Don't allow assignment of objects.
   */
  SCALESTACK_LOCAL
  Option& operator=(const Option&);

  /**
   * Set the help string of the option.
   */
  SCALESTACK_LOCAL
  void _setHelp(const std::string& help);

  /**
   * Set the value name of the option.
   */
  SCALESTACK_LOCAL
  void _setValueName(const std::string& value_name);

  /**
   * Set the default value of the option.
   */
  SCALESTACK_LOCAL
  void _setDefaultValue(const std::string& default_value);

  /**
   * Get the value of the option.
   */
  SCALESTACK_LOCAL
  void _setValue(const std::string& value, bool overwrite);

  /**
   * Set whether this is a valid option or not.
   */
  SCALESTACK_LOCAL
  void _setValid(bool valid);

  /**
   * Print a help message for this option.
   */
  SCALESTACK_LOCAL
  void _printHelp(void);

  bool _is_valid;
  bool _is_set;
  Module& _module;
  const std::string _name;
  std::string _help;
  std::string _value_name;
  std::string _default_value;
  std::string _value;
};

} /* namespace Kernel */

} /* namespace ScaleStack */

#endif /* SCALESTACK_KERNEL_OPTION_H */
