/*
 * Scale Stack
 *
 * Copyright (C) 2010 Eric Day (eday@oddments.org)
 * All rights reserved.
 *
 * Use and distribution licensed under the BSD license. See the
 * COPYING file in the root project directory for full text.
 */

/**
 * @file
 * @brief TCP Socket Connection Definitions
 */

#include "config.h"

#include <cerrno>
#include <cstring>
#include <sys/types.h> /* Must be before TCP headers for FreeBSD. */
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <sys/socket.h>
#include <unistd.h>

#include <ScaleStack/Kernel/Module.h>
#include <ScaleStack/SocketTCP/Connection.h>
#include <ScaleStack/SocketTCP/ConnectionService.h>

using namespace std;

namespace ScaleStack
{

namespace SocketTCP
{

/*
 * Public methods.
 */

Connection::Connection(ConnectionService& connection_service):
  Event::Handler(connection_service),
  _module(connection_service.getModule()),
  _connecting(false),
  _waiting_for_consume(false),
  _call_flush_write(false),
  _shutdown_write(false),
  _connection_dead(false),
  _file_descriptor(-1),
  _read_buffer_offset(),
  _read_buffer_size(),
  _write_buffer_offset(),
  _write_buffer_size(),
  _connect_addrinfo_list(),
  _connect_addrinfo()
{
}

Connection::~Connection()
{
  if (_file_descriptor != -1)
    close(_file_descriptor);

  if (_connect_addrinfo_list != NULL)
    freeaddrinfo(_connect_addrinfo_list);
}

void Connection::setFileDescriptor(int file_descriptor)
{
  _setFileDescriptor(file_descriptor);
  watchRead();
}

void Connection::connect(const std::string& host, const std::string& port)
{
  struct addrinfo addrinfo_options;

  if (_connect_addrinfo_list != NULL)
  {
    freeaddrinfo(_connect_addrinfo_list);
    _connect_addrinfo_list = NULL;
  }

  memset(&addrinfo_options, 0, sizeof(struct addrinfo));
  addrinfo_options.ai_socktype = SOCK_STREAM;
  addrinfo_options.ai_protocol = IPPROTO_TCP;

  int return_code = getaddrinfo(host.c_str(), port.c_str(), &addrinfo_options,
                                &_connect_addrinfo_list);
  if (return_code != 0)
  {
    _module.logError("Failed getaddrinfo: %s", gai_strerror(return_code));
    error();
    return;
  }

  reconnect();
}

void Connection::reconnect(void)
{
  _setFileDescriptor(-1);
  _connect_addrinfo = _connect_addrinfo_list;
  _continueConnect();
}

void Connection::consume(size_t size)
{
  if (size == 0)
    return;

  _consume(size);
  watchRead();
}

void Connection::readEOF(void)
{
  error();
}

size_t Connection::write(uint8_t* buffer, size_t size, bool flush)
{
  if (size == 0 || _connecting)
    return 0;

  size_t written;

  /* Clear this because a forced write could remove the need for flush. */
  _call_flush_write = false;

  if (_write_buffer_size == 0 && (flush || size >= _write_buffer_max_size))
    written = 0;
  else if (size <= (_write_buffer_max_size - (_write_buffer_offset + _write_buffer_size)))
  {
    /* Everything fits in the write buffer, so copy it in there. */
    memcpy(_write_buffer + _write_buffer_offset + _write_buffer_size,
           buffer,
           size);
    _write_buffer_size += size;
    if (flush)
      _flush();

    return size;
  }
  else
  {
    /* Not everything could fit in the buffer. */
    if (size < _write_buffer_max_size)
    {
      /* Copy as much as we can in if the requested amount isn't too large. */
      written = _write_buffer_max_size - (_write_buffer_offset + _write_buffer_size);
      if (written > 0)
      {
        memcpy(_write_buffer + _write_buffer_offset + _write_buffer_size,
               buffer,
               written);
        _write_buffer_size += written;
      }
    }
    else
      written = 0;

    _flush();
  }

  if (_write_buffer_size == 0 &&
      (flush || (size - written) >= _write_buffer_max_size))
  {
    /*
     * There is nothing in the buffer and either a flush was requested or
     * the size is larger than the buffer, so just try writing directly.
     */
    written += _write(buffer + written, size - written);
    if (written == size)
      return size;
  }

  /* If the remaining is larger than the buffer, don't bother buffering it. */
  if ((size - written) >= _write_buffer_max_size)
  {
    _call_flush_write = true;
    return written;
  }

  size_t fill_size = _write_buffer_max_size - (_write_buffer_offset + _write_buffer_size);
  if (fill_size == 0)
  {
    _call_flush_write = true;
    return written;
  }

  if ((size - written) <= fill_size)
    fill_size = size - written;
  else
    _call_flush_write = true;

  /* Fill the buffer with what we can. */
  memcpy(_write_buffer + _write_buffer_offset + _write_buffer_size,
         buffer + written,
         fill_size);
  _write_buffer_size += fill_size;

  return written + fill_size;
}

void Connection::shutdownWrite(void)
{
  _shutdown_write = true;
  runNow();
}

void Connection::connected(void)
{
}

void Connection::run(void)
{
  if (_shutdown_write && _write_buffer_size == 0)
  {
    int return_code = shutdown(_file_descriptor, SHUT_WR);
    if (return_code == -1)
    {
      _module.logError("Could not shut down write socket: %d", errno);
      error();
      return;
    }
  }
}

/*
 * Private methods.
 */

void Connection::_setFileDescriptor(int file_descriptor)
{
  if (_file_descriptor != -1)
    close(_file_descriptor);

  _file_descriptor = file_descriptor;

  if (_file_descriptor != -1)
  {
    int return_code = 1;
    return_code = setsockopt(_file_descriptor, IPPROTO_TCP, TCP_NODELAY,
                             &return_code, static_cast<socklen_t>(sizeof(int)));
    if (return_code == -1 && errno != EOPNOTSUPP)
    {
      _module.logError("Could not set TCP_NODELAY: %d", errno);
      error();
      return;
    }
  }

  Event::Handler::setFileDescriptor(file_descriptor);
}

void Connection::_continueConnect(void)
{
  while (true)
  {
    if (_file_descriptor == -1)
    {
      int file_descriptor = socket(_connect_addrinfo->ai_family,
                                   _connect_addrinfo->ai_socktype,
                                   _connect_addrinfo->ai_protocol);
      if (file_descriptor == -1)
      {
        _module.logError("Failed socket: %d", errno);
        error();
        return;
      }

      _setFileDescriptor(file_descriptor);
    }

    int return_code = ::connect(_file_descriptor,
                                _connect_addrinfo->ai_addr,
                                _connect_addrinfo->ai_addrlen);
    if (return_code == 0 || errno == EISCONN)
    {
      _module.logInfo("Connected");
      watchRead();
      connected();
      return;
    }

    if (errno == EAGAIN || errno == EINTR)
      continue;

    if (errno == EINPROGRESS || errno == EALREADY)
    {
      _module.logDebug("Connection in progress");
      _connecting = true;
      watchWrite();
      return;
    }

    if (errno == ECONNREFUSED || errno == ENETUNREACH || errno == ETIMEDOUT)
    {
      _connect_addrinfo = _connect_addrinfo->ai_next;
      if (_connect_addrinfo == NULL)
      {
        _module.logError("Could not connect");
        error();
        return;
      }

      _setFileDescriptor(-1);
      continue;
    }

    _module.logError("Failed connect: %d", errno);
    error();
    return;
  }
}

void Connection::readReady(int)
{
  while (!_connecting && _read_buffer_size != _read_buffer_max_size)
  {
    size_t free_size = _read_buffer_max_size - (_read_buffer_offset + _read_buffer_size);
    if (free_size <= (_read_buffer_max_size / 4))
    {
      /*
       * We don't want to move any buffer space if the derived class could
       * not consume it all because it probably has active pointers into
       * this buffer.
       */
      if (_waiting_for_consume)
        return;

      memmove(_read_buffer,
              _read_buffer + _read_buffer_offset,
              _read_buffer_size);
      _read_buffer_offset = 0;
      free_size = _read_buffer_max_size - _read_buffer_size;
    }

    ssize_t return_size = ::read(_file_descriptor,
                                 _read_buffer + _read_buffer_offset + _read_buffer_size,
                                 free_size);
    if (return_size == 0)
    {
      _module.logInfo("Received EOF on read");
      _shutdown_write = false;

      if (_write_buffer_size > 0)
        _flush();

      if (_write_buffer_size > 0 || _call_flush_write)
      {
        _connection_dead = true;
        return;
      }

      readEOF();
      return;
    }
    else if (return_size == -1)
    {
      if (errno == EAGAIN)
      {
        watchRead();
        return;
      }
      else if (errno == EINTR)
        continue;

      _module.logInfo("Received error on read: %d", errno);
      _shutdown_write = false;
      error();
      return;
    }

    _module.logDebug("Read %d bytes", return_size);

    _read_buffer_size += return_size;
    _consume(read(_read_buffer + _read_buffer_offset, _read_buffer_size));
  }
}

void Connection::writeReady(int)
{
  if (_connecting)
  {
    _connecting = false;
    _continueConnect();
    return;
  }

  if (_write_buffer_size != 0)
    _flush();

  if (_write_buffer_size == 0)
  {
    if (_call_flush_write)
    {
      _call_flush_write = false;
      flushWrite();

      if (!_call_flush_write)
        watchRead();
    }

    if (_connection_dead && _write_buffer_size == 0 && !_call_flush_write)
      error();
  }
}

void Connection::_consume(size_t size)
{
  if (size == 0)
  {
    _waiting_for_consume = true;
    return;
  }

  if (size > _read_buffer_size)
  {
    _module.logError("Consumed more data than is in buffer: %u > %u",
                     size, _read_buffer_size);
    error();
  }
  else if (size == _read_buffer_size)
  {
    _read_buffer_size = 0;
    _read_buffer_offset = 0;
    _waiting_for_consume = false;
  }
  else
  {
    _read_buffer_size -= size;
    _read_buffer_offset += size;
    _waiting_for_consume = true;
  }
}

void Connection::_flush(void)
{
  size_t return_size = _write(_write_buffer + _write_buffer_offset,
                              _write_buffer_size);
  if (return_size == 0)
    return;

  if (return_size == _write_buffer_size)
  {
    _write_buffer_size = 0;
    _write_buffer_offset = 0;

    if (_shutdown_write)
      runNow();
  }
  else
  {
    _write_buffer_size -= return_size;
    _write_buffer_offset += return_size;

    if ((_write_buffer_offset + _write_buffer_size) >= (_write_buffer_max_size / 2))
    {
      memmove(_write_buffer,
              _write_buffer + _write_buffer_offset,
              _write_buffer_size);
      _write_buffer_offset = 0;
    }
  }
}

size_t Connection::_write(uint8_t* buffer, size_t size)
{
  while (true)
  {
    ssize_t return_size = ::write(_file_descriptor, buffer, size);
    if (return_size == 0)
    {
      _module.logInfo("Received EOF on write");
      _shutdown_write = false;
      error();
      return 0;
    }
    else if (return_size == -1)
    {
      if (errno == EAGAIN)
      {
        watchWrite();
        return 0;
      }
      else if (errno == EINTR)
        continue;

      _module.logInfo("Received error on write: %d", errno);
      _shutdown_write = false;
      error();
      return 0;
    }

    _module.logDebug("Wrote %d bytes", return_size);

    if (static_cast<size_t>(return_size) != size)
      watchWrite();

    return return_size;
  }
}

} /* namespace SocketTCP */

} /* namespace ScaleStack */
