/*
 * Scale Stack
 *
 * Copyright (C) 2010 Eric Day (eday@oddments.org)
 * All rights reserved.
 *
 * Use and distribution licensed under the BSD license. See the
 * COPYING file in the root project directory for full text.
 */

/**
 * @file
 * @brief TCP Socket Connection Declarations
 */

#ifndef SCALESTACK_SOCKETTCP_CONNECTION_H
#define SCALESTACK_SOCKETTCP_CONNECTION_H

#include <netdb.h>
#include <stdint.h>
#include <string>
#include <sys/socket.h>
#include <sys/types.h>

#include <ScaleStack/Common/Macros.h>
#include <ScaleStack/Event/Handler.h>

namespace ScaleStack
{

namespace Kernel
{

class Module;

} /* namespace Kernel */

namespace SocketTCP
{

class ConnectionService;

class SCALESTACK_API Connection: public Event::Handler
{
public:

  Connection(ConnectionService& connection_service);

  virtual ~Connection();

  /**
   * Set the file descriptor to use for this connection. This is usually
   * the result of an accepted connection.
   *
   * @param[in] file_descriptor File descriptor to set.
   */
  void setFileDescriptor(int file_descriptor);

  /**
   * Connect to a remote host.
   *
   * @param[in] host Host to connect to.
   * @param[in] port Port to connect to.
   */
  void connect(const std::string& host, const std::string& port);

  /**
   * Reconnect using the previous connect host and port.
   */
  void reconnect(void);

  /**
   * This is called when data has been read and needs to be handled by
   * derived classes.
   *
   * @param[in] buffer Data that was read.
   * @param[in] size Size of data available in buffer.
   * @return Amount of data that was consumed.
   */
  virtual size_t read(uint8_t* buffer, size_t size) = 0;

  /**
   * Notify the connection that more data from the read buffer has been
   * consumed. This is required after a read() call was not able to consume
   * all the data. This can be called at a later time to remove the data
   * from the read buffer, such as in flushWrite(). If the read buffer was
   * previously full the connection will start watching for new data.
   *
   * @param[in] size The amount of data to consume.
   */
  void consume(size_t size);

  /**
   * Notify the connection that EOF has been encountered while reading. THe
   * default behavior is to call error().
   */
  virtual void readEOF(void);

  /**
   * Write data to the connection. By default this is buffered and not
   * flushed, but it will be flushed if the buffer becomes full. If the
   * buffer being written is large enough, it will be written directly to
   * the connection without being buffered. If this method is not able to
   * write or buffer all data, the connection will return with the partial
   * amount and call flushWrite() when it is ready again.
   *
   * @param[in] buffer Buffer to write.
   * @param[in] size Size of buffer.
   * @param[in] flush Whether to flush data immediately or not.
   * @return Amount of data written or buffered.
   */
  size_t write(uint8_t* buffer, size_t size, bool flush = false);

  /**
   * This is called to notify derived classes that they should start flushing
   * write data again by calling write(). This is called when the connection
   * is ready for writing after a call to write() did not write all data.
   */
  virtual void flushWrite(void) = 0;

  /**
   * Shutdown the write end of the socket.
   */
  void shutdownWrite(void);

  /**
   * This is called to notify derived classes that a connection has been
   * established. The default behavior is to ignore.
   */
  virtual void connected(void);

  /**
   * Handle various events for this connection.
   */
  virtual void run(void);

protected:

  Kernel::Module& _module;

private:

  /**
   * Don't allow copying of objects.
   */
  SCALESTACK_LOCAL
  Connection(const Connection&);

  /**
   * Don't allow assignment of objects.
   */
  SCALESTACK_LOCAL
  Connection& operator=(const Connection&);

  /**
   * Set file descriptor being used. This can be either from an accept
   * or connect.
   */
  SCALESTACK_LOCAL
  void _setFileDescriptor(int file_descriptor);

  /**
   * Try connecting to the next entry in the address info list.
   */
  SCALESTACK_LOCAL
  void _continueConnect(void);

  /**
   * This will read data from the connection into the buffer and call
   * read(). This implements a method in the base class so does not have
   * the _ prefix.
   *
   * @param[in] file_descriptor File descriptor to read data on.
   */
  void readReady(int file_descriptor);

  /**
   * This flushes any data that was in the buffer and calls writeReady()
   * for derived classes if needed. This implements a method in the base
   * class so does not have the _ prefix.
   *
   * @param[in] file_descriptor File descriptor to write data to.
   */
  void writeReady(int file_descriptor);

  /**
   * Perform the read buffer consumption, used by both readReady() and
   * consume().
   *
   * @param[in] size Amount of buffer to consume.
   */
  SCALESTACK_LOCAL
  void _consume(size_t size);

  /**
   * Flush data that is in the write buffer.
   */
  SCALESTACK_LOCAL
  void _flush(void);

  /**
   * Write data to connection and handle any errors.
   *
   * @param[in] buffer Buffer to write.
   * @param[in] size Size of buffer.
   * @return Amount of data written.
   */
  SCALESTACK_LOCAL
  size_t _write(uint8_t* buffer, size_t size);

  static const size_t _read_buffer_max_size = 32768;
  static const size_t _write_buffer_max_size = 32768;

  bool _connecting;
  bool _waiting_for_consume;
  bool _call_flush_write;
  bool _shutdown_write;
  bool _connection_dead;
  int _file_descriptor;
  size_t _read_buffer_offset;
  size_t _read_buffer_size;
  size_t _write_buffer_offset;
  size_t _write_buffer_size;
  struct addrinfo* _connect_addrinfo_list;
  struct addrinfo* _connect_addrinfo;
  uint8_t _read_buffer[_read_buffer_max_size];
  uint8_t _write_buffer[_write_buffer_max_size];
};

} /* namespace SocketTCP */

} /* namespace ScaleStack */

#endif /* SCALESTACK_SOCKETTCP_CONNECTION_H */
