/*
 * Scale Stack
 *
 * Copyright (C) 2010 Eric Day (eday@oddments.org)
 * All rights reserved.
 *
 * Use and distribution licensed under the BSD license. See the
 * COPYING file in the root project directory for full text.
 */

/**
 * @file
 * @brief TCP Socket Connection Service Declarations
 */

#ifndef SCALESTACK_SOCKETTCP_CONNECTIONSERVICE_H
#define SCALESTACK_SOCKETTCP_CONNECTIONSERVICE_H

#include <string>

#include <ScaleStack/Common/Macros.h>
#include <ScaleStack/Event/Service.h>

namespace ScaleStack
{

namespace SocketTCP
{

class Connection;

class SCALESTACK_API ConnectionService: public Event::Service
{
public:

  ConnectionService(Kernel::Module& module);

  virtual ~ConnectionService();

  /**
   * Get a new TCP socket connection from a derived class.
   *
   * @return Connection that was added.
   */
  virtual Connection* addConnection(void) = 0;

  /**
   * Remove a TCP socket connection that was added in addConnection().
   *
   * @param[in] connection The connection to remove.
   */
  virtual void removeConnection(Connection* connection) = 0;

  /**
   * Add a new connection that has been accepted.
   *
   * @param[in] file_descriptor File descriptor for the new connection.
   * @return Connection that was added.
   */
  Connection* addAccepted(int file_descriptor);

  /**
   * Add a new connection and connect to the given host:port.
   *
   * @param[in] host Host to connect to.
   * @param[in] port Port to connect to.
   * @return Connection that was added.
   */
  Connection* addConnect(const std::string& host, const std::string& port);

private:

  /**
   * Don't allow copying of objects.
   */
  SCALESTACK_LOCAL
  ConnectionService(const ConnectionService&);

  /**
   * Don't allow assignment of objects.
   */
  SCALESTACK_LOCAL
  ConnectionService& operator=(const ConnectionService&);

  /**
   * Add a new connection to use for accept() or connect().
   */
  SCALESTACK_LOCAL
  Connection* _addConnection(void);

  /**
   * See Event::Service::removeHandler().
   */
  void removeHandler(Event::Handler* handler);
};

} /* namespace SocketTCP */

} /* namespace ScaleStack */

#endif /* SCALESTACK_SOCKETTCP_CONNECTIONSERVICE_H */
