/*
 * Scale Stack
 *
 * Copyright (C) 2010 Eric Day (eday@oddments.org)
 * All rights reserved.
 *
 * Use and distribution licensed under the BSD license. See the
 * COPYING file in the root project directory for full text.
 */

/**
 * @file
 * @brief TCP Socket Listener Definitions
 */

#include "config.h"

#include <cerrno>
#include <exception>
#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>

#include <ScaleStack/Kernel/Module.h>
#include <ScaleStack/Kernel/Option.h>
#include <ScaleStack/SocketTCP/Connection.h>
#include <ScaleStack/SocketTCP/ConnectionService.h>
#include <ScaleStack/SocketTCP/Listener.h>
#include <ScaleStack/SocketTCP/ListenerService.h>

using namespace std;

namespace ScaleStack
{

namespace SocketTCP
{

/*
 * Public methods.
 */

Listener::Listener(ListenerService& listener_service,
                   ConnectionService& connection_service,
                   struct addrinfo* addrinfo):
  Event::Handler(listener_service),
  _file_descriptor(-1),
  _module(listener_service.getModule()),
  _connection_service(connection_service)
{
  char host_name[NI_MAXHOST];
  char port_name[NI_MAXSERV];
  int return_code;

  return_code = getnameinfo(addrinfo->ai_addr,
                            addrinfo->ai_addrlen,
                            host_name, NI_MAXHOST,
                            port_name, NI_MAXSERV,
                            NI_NUMERICHOST | NI_NUMERICSERV);
  if (return_code != 0)
  {
    _module.logFatal("Error adding TCP listener, getnameinfo: %s",
                     gai_strerror(return_code));
  }

  _module.logInfo("Trying to listen on %s:%s", host_name, port_name);

  _file_descriptor = socket(addrinfo->ai_family,
                            addrinfo->ai_socktype,
                            addrinfo->ai_protocol);
  if (_file_descriptor == -1)
    _module.logFatal("Error adding TCP listener, socket: %d", errno);

  int option = 1;
  return_code = setsockopt(_file_descriptor, SOL_SOCKET, SO_REUSEADDR,
                           &option, sizeof(option));
  if (return_code == -1)
  {
    close(_file_descriptor);
    _module.logFatal("Error adding TCP listener, setsockopt: %d", errno);
  }

  return_code = ::bind(_file_descriptor,
                       addrinfo->ai_addr,
                       addrinfo->ai_addrlen);
  if (return_code == -1)
  {
    close(_file_descriptor);
    if (errno == EADDRINUSE)
    {
      _module.logInfo("Bind failed, address already in use");
      throw exception();
    }
    _module.logFatal("Error adding TCP listener, bind: %d", errno);
  }

  return_code = listen(_file_descriptor,
                       _module.getOption("backlog").getIntValue());
  if (return_code == -1)
  {
    close(_file_descriptor);
    _module.logFatal("Error adding TCP listener, listen: %d", errno);
  }

  try
  {
    _service.addHandler(this);
    setFileDescriptor(_file_descriptor);
    watchRead();
  }
  catch (exception&)
  {
    close(_file_descriptor);
    throw;
  }

  _module.logNotice("Listening on %s:%s", host_name, port_name);

  _module.logInfo("Listener constructed");
}

Listener::~Listener()
{
  close(_file_descriptor);
  _module.logInfo("Listener destructed");
}

void Listener::readReady(int)
{
  while (true)
  {
    char host_name[NI_MAXHOST];
    char port_name[NI_MAXSERV];
    struct sockaddr sockaddr;
    socklen_t sockaddr_length = sizeof(struct sockaddr);

    int accepted_file_descriptor = accept(_file_descriptor,
                                          &sockaddr,
                                          &sockaddr_length);
    if (accepted_file_descriptor == -1)
    {
      if (errno == EAGAIN)
      {
        watchRead();
        return;
      }
      else if (errno == EINTR)
        continue;
      else if (errno == EMFILE)
      {
        _module.logError("Accept failed due to too many open files");
        watchRead();
        return;
      }

      _module.logFatal("Accept failed: %d", errno);
    }

    int return_code = getnameinfo(&sockaddr,
                                  sockaddr_length,
                                  host_name, NI_MAXHOST,
                                  port_name, NI_MAXSERV,
                                  NI_NUMERICHOST | NI_NUMERICSERV);
    if (return_code != 0)
    {
      _module.logError("Error accepting TCP connection, getnameinfo: %s",
                       gai_strerror(return_code));
    }
    else
      _module.logNotice("Accepted %s:%s", host_name, port_name);

    try
    {
      _connection_service.addAccepted(accepted_file_descriptor);
    }
    catch (exception& e)
    {
      _module.logError("Error adding TCP connection: %s", e.what());
      close(accepted_file_descriptor);
    }
  }
}

} /* namespace SocketTCP */

} /* namespace ScaleStack */
