/*
 * Scale Stack
 *
 * Copyright (C) 2010 Eric Day (eday@oddments.org)
 * All rights reserved.
 *
 * Use and distribution licensed under the BSD license. See the
 * COPYING file in the root project directory for full text.
 */

/**
 * @file
 * @brief TCP Socket Listener Service Definitions
 */

#include "config.h"

#include <cstring>
#include <exception>
#include <netdb.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/types.h>

#include <ScaleStack/Kernel/Module.h>
#include <ScaleStack/Kernel/Option.h>
#include <ScaleStack/SocketTCP/Listener.h>
#include <ScaleStack/SocketTCP/ListenerService.h>

using namespace std;

namespace ScaleStack
{

namespace SocketTCP
{

/*
 * Public Functions.
 */

void ListenerOptions(Kernel::Module& module, const char* port)
{
  module.addOption("hosts", "List of host/port pairs listen on.",
                   "HOST[:PORT][,HOST[:PORT]]...", "");
  module.addOption("ports", "List of ports to listen on.", "PORT[,PORT]...",
                   port);
  module.addOption("backlog", "Backlog value to use for listen().", "BACKLOG",
                   "64");
}

/*
 * Public methods.
 */

ListenerService::ListenerService(Kernel::Module& module,
                                 ConnectionService& connection_service):
  Event::Service(module)
{
  vector<pair<string,string> > hosts = module.getOption("hosts").getHostListValue();
  vector<string> ports = module.getOption("ports").getListValue();

  try
  {
    for (vector<pair<string,string> >::iterator host = hosts.begin();
         host != hosts.end();
         ++host)
    {
      if (host->second == "")
      {
        for (vector<string>::iterator port = ports.begin();
             port != ports.end();
             ++port)
        {
          _addListeners(connection_service, host->first, *port);
        }
      }
      else
        _addListeners(connection_service, host->first, host->second);
    }
  }
  catch (exception&)
  {
    shutdown();
    throw;
  }
}

ListenerService::~ListenerService()
{
  shutdown();
}

/*
 * Private methods.
 */

void ListenerService::_addListeners(ConnectionService& connection_service,
                                    std::string host,
                                    std::string port)
{
  struct addrinfo addrinfo_options;
  struct addrinfo *addrinfo_list;
  struct addrinfo *addrinfo;

  memset(&addrinfo_options, 0, sizeof(struct addrinfo));
  addrinfo_options.ai_flags  = AI_PASSIVE;
  addrinfo_options.ai_family = AF_UNSPEC;
  addrinfo_options.ai_socktype = SOCK_STREAM;
  addrinfo_options.ai_protocol = IPPROTO_TCP;

  int return_code = getaddrinfo(host == "" ? NULL : host.c_str(),
                                port.c_str(),
                                &addrinfo_options,
                                &addrinfo_list);
  if (return_code != 0)
  {
    _module.logFatal("Error adding TCP listeners, getaddrinfo: %s",
                     gai_strerror(return_code));
  }

  
  /* We only require one host:port pair from the list to succeed. */
  bool added = false;
  for (addrinfo = addrinfo_list; addrinfo != NULL; addrinfo = addrinfo->ai_next)
  {
    try
    {
      new Listener(*this, connection_service, addrinfo);
      added = true;
    }
    catch (exception&)
    {
    }
  }

  freeaddrinfo(addrinfo_list);

  if (!added)
  {
    _module.logFatal("Could not load any TCP listeners for %s:%s",
                     host.c_str(), port.c_str());
  }
}

void ListenerService::removeHandler(Event::Handler* handler)
{
  delete handler;
}

} /* namespace SocketTCP */

} /* namespace ScaleStack */
